## "DEPENDENCIES REQUIRED: PTSA, matplotlib, scikit-learn, CMLReaders. ##
import cluster_helper.cluster
import os, sys, glob
import numpy.testing
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import pandas as pd
from ptsa.data.readers import BaseEventReader
from ptsa.data.readers import TalReader
from ptsa.data.readers import EEGReader
from ptsa.data.filters import MonopolarToBipolarMapper
from ptsa.data.filters import ButterworthFilter
from ptsa.data.filters.MorletWaveletFilter import MorletWaveletFilter
from ptsa.data.readers import JsonIndexReader
import xarray as xr
import sklearn
from sklearn.linear_model import LogisticRegression
from sklearn.decomposition import FastICA, PCA
from sklearn.metrics import roc_auc_score, roc_curve
from scipy.stats.mstats import zscore
from scipy import stats
import json
import math
import scipy
import random
from ptsa.data.TimeSeriesX import TimeSeriesX
lr_classifier = LogisticRegression(C=7.2e-4, penalty='l2', class_weight='balanced', solver='liblinear')
from cmlreaders import CMLReader

    ############################################     CREATE DATA STRUCTURES CODES   ##################################################
def get_save_data_cml(sub, outpath):
    """Returns most relevant information for a subject (output variables are described in beginning of code). Output variables are stored as lists, with the ith element corresponding to the ith session. The output is a dictionary.

    INPUTS: 'sub' is a string of the subject name (eg 'R1001P')
    'outpath' is where you want the data to be saved to"""

#INITIALIZE OUTPUT VARIABLES
    events_out = [] # Event structure for word events
    exps_out = [] # Successful experiments
    sessions_out = [] # List of sessions
    contacts_out = [] # List of channels
    num_contacts_out = [] # Number of channels
    pairs_out = [] # List of bipolar pairs
    num_pairs_out = [] # Number of pairs
    recalls_out = [] # Recalls information (0 = not recalled, 1 = recalled)
    rec_out = [] # Indices of recalled words
    not_rec_out = [] # Indices of not recalled words
    lists_out = [] # Array of the number of word lists
    num_lists_out = [] # Number of lists per session
    word_lists_out = [] # Array matching each epoch to a word list
    num_words_out = [] # Number of words
    powers_car_uni_out = [] # Monopolar power, univariate [epXelec(8 entries per elec for 8 power bins)]
    powers_pairs_uni_out = [] # Bipolar power, univariate [epXelec(8 entries per elec for 8 power bins)]
    powers_car_multi_out = [] # Monopolar power, multivariate [epXelec(8 entries per elec for 8 power bins)]
    powers_pairs_multi_out = [] # Bipolar power, multivariate [epXelec(8 entries per elec for 8 power bins)]
    S = [] # Strip electrodes
    G = [] # Grid electrodes
    D = [] # Depth electrodes
    stds_car_out = [] # Standard deviation of monopolar data [epXelec]
    stds_pairs_out = [] # Standard deviation of bipolar data [epXelec]
    kurs_car_out = [] # Kurtosis of monopolar data [epXelec]
    kurs_pairs_out = [] # Kurtosis of bipolar data [epXelec]

    # INITIALIZE PARAMETERS
    buffer = 1.5 # BUFFER USED IN MORELET WAVELET CONVOLUTION TO CONTROL FOR EDGE EFFECTS
    f_min_uni = 70 # Univariate analyses min frequency
    f_max_uni = 200 # Univariate analyses max frequency
    f_min_multi = 3 # Multivariate analyses min frequency
    f_max_multi = 175 # Multivariate analyses max frequency
    num_freqs = 8 # Number of frequency bins for power analysis
    exps = ['FR1', 'catFR1'] # Experiments to load
    
    ix = CMLReader.get_data_index("r1") #Obtain CML reader index
    sub_exp = ix[ix["subject"] == sub]["experiment"] #Get available experiments for the subject
    num_catfr1 = np.size(np.where(sub_exp=='catFR1')[0]) #Number of catFR1
    num_fr1 = np.size(np.where(sub_exp=='FR1')[0]) #Number of FR1

    for exp,num_sessions in zip(exps,[num_fr1, num_catfr1]):
        for sess in range(num_sessions):
            try:
                print(sess)
                reader = CMLReader(sub, exp, sess)
                contacts = reader.load("contacts") #Get monopolar contacts
                num_contacts = len(contacts)               
                pairs = reader.load("pairs") #Get bipolar pairs                
                num_pairs = len(pairs)               

                events = reader.load("events") #Get epoch information
                words = events[events["type"] == "WORD"]
                
                recall_info = get_recalls(words) # Get recall information

                ts_car = reader.load_eeg(events=words, rel_start=0, rel_stop=1600, scheme=contacts) #Load the CAR timeseries
                # Make a dummy xarray object to borrow coordinate structure from and create the timeseries object
                dummy = xr.DataArray([[[0,0],[0,0],[0,0]]], coords=[[0], [0,0,0], [0,0]], dims=['events', 'channels', 'time'])
                coords=dummy.coords
                coords['channels'] = range(ts_car.shape[1])
                coords['events'] = range(ts_car.shape[0])
                coords['time'] = range(ts_car.shape[2])
                coords['samplerate']=ts_car.samplerate
                ts_car = TimeSeriesX(ts_car.data, coords=coords)
                # Filter the timeseries at line noise harmonics
                raw_car = ButterworthFilter(time_series=ts_car, freq_range=[58., 62.], filt_type='stop',
                                                        order=4).filter()
                raw_car = ButterworthFilter(time_series=raw_car, freq_range=[118., 122.], filt_type='stop',
                                                                order=4).filter()
                filt_car = ButterworthFilter(time_series=raw_car, freq_range=[178., 182.], filt_type='stop',
                                                                order=4).filter()
                ts_car_buffer = filt_car.add_mirror_buffer(1.5)
                #Repeat the above steps for bipolar data
                ts_pairs = reader.load_eeg(events=words, rel_start=0, rel_stop=1600, scheme=pairs)
                dummy = xr.DataArray([[[0,0],[0,0],[0,0]]], coords=[[0], [0,0,0], [0,0]], dims=['events', 'bipolar_pairs', 'time'])
                coords=dummy.coords
                coords['bipolar_pairs'] = range(ts_pairs.shape[1])
                coords['events'] = range(ts_pairs.shape[0])
                coords['time'] = range(ts_pairs.shape[2])
                coords['samplerate']=ts_pairs.samplerate
                ts_pairs = TimeSeriesX(ts_pairs.data, coords=coords)
                raw_pairs = ButterworthFilter(time_series=ts_pairs, freq_range=[58., 62.], filt_type='stop',
                                                        order=4).filter()
                raw_pairs = ButterworthFilter(time_series=raw_pairs, freq_range=[118., 122.], filt_type='stop',
                                                                order=4).filter()
                filt_pairs = ButterworthFilter(time_series=raw_pairs, freq_range=[178., 182.], filt_type='stop',
                                                                order=4).filter()
                ts_pairs_buffer = filt_pairs.add_mirror_buffer(1.5)

                # Calculate the standard deviation and bipolar pairs
                stds_car = np.std(filt_car.values, axis=2)
                kurs_car =  scipy.stats.kurtosis(filt_car.values, axis=2, fisher=True)
                stds_pairs = np.std(filt_pairs.values, axis=2)
                kurs_pairs =  scipy.stats.kurtosis(filt_pairs.values, axis=2, fisher=True)
                
                # CALCULATE POWERS (uni = univariate, multi = multivariate)
                powers_car_uni = get_powers(ts_car_buffer, 'M', buffer, f_min_uni, f_max_uni, num_freqs)
                powers_pairs_uni = get_powers(ts_pairs_buffer, 'B', buffer, f_min_uni, f_max_uni, num_freqs)
                powers_car_multi = get_powers(ts_car_buffer, 'M', buffer, f_min_multi, f_max_multi, num_freqs)
                powers_pairs_multi = get_powers(ts_pairs_buffer, 'B', buffer, f_min_multi, f_max_multi, num_freqs)
                
                # Append to output variables that are eventually saved out
                exps_out.append(exp)
                sessions_out.append(sess)
                events_out.append(words)

                contacts_out.append(contacts)
                num_contacts_out.append(num_contacts)
                pairs_out.append(pairs)
                num_pairs_out.append(num_pairs)
                S.append(np.size(np.where(contacts.type=='S')[0]))
                G.append(np.size(np.where(contacts.type=='G')[0]))
                D.append(np.size(np.where(contacts.type=='D')[0]))

                num_words_out.append(len(words))
                recalls_out.append(recall_info['recalls']) # 0 - Not recalled, 1 - recalled
                rec_out.append(recall_info['rec']) # Indicies of recalled words
                not_rec_out.append(recall_info['not_rec']) # Indicies of non-recalled words
                word_lists_out.append(recall_info['word_lists']) # Indiciation of what words belong to which lists
                lists_out.append(recall_info['lists']) # List of word lists
                num_lists_out.append(recall_info['num_lists']) # Number of lists

                stds_car_out.append(stds_car)
                kurs_car_out.append(kurs_car)
                stds_pairs_out.append(stds_pairs)
                kurs_pairs_out.append(kurs_pairs)

                powers_car_uni_out.append(powers_car_uni)
                powers_pairs_uni_out.append(powers_pairs_uni)
                powers_car_multi_out.append(powers_car_multi)
                powers_pairs_multi_out.append(powers_pairs_multi)
                #break

            except Exception as e:
                # Return error information
                print(e)
                np.savez(outpath+'/'+sub+'_'+exp+'_'+str(sess)+'_error', e)
            
    data = {'events': events_out, 'exps': exps_out, 'sessions':sessions_out, 
                        'contacts':contacts_out, 'num_contacts': num_contacts_out, 'pairs':pairs_out, 'num_pairs': num_pairs_out,
                        'recalls': recalls_out,'rec': rec_out, 'not_rec': not_rec_out, 'lists': lists_out, 'num_lists': num_lists_out, 
                        'word_lists': word_lists_out, 'num_words_sess': num_words_out, 'powers_car_uni': 
                        powers_car_uni_out, 'powers_pairs_uni': powers_pairs_uni_out, 'powers_car_multi': powers_car_multi_out, 
                        'powers_pairs_multi': powers_pairs_multi_out, 'S':S, 'G':G, 'D':D, 'stds_car':stds_car_out,
                        'stds_pairs':stds_pairs_out, 'kurs_car':kurs_car_out, 'kurs_pairs': kurs_pairs_out}
    np.savez(outpath+'/'+sub+'_data', data=data)


def get_recalls(words):

    """GET RECALL INFORMATION. 

    Inputs: 'words' is epoch information generated by the CMLReader

    Outputs: 'recalls' is the vector(s) of recall results for the experiment's session(s). 'Rec' and 'Not_rec' are objects that give indicies for each session when a word was recalled or not recalled. 'lists' is a list of all of the unique word_lists for the session. 'num_lists' is the number of lists per session. 'word_lists' is a vector which associates each epoch to its respective word list. 'offsets' describe when (in terms of data points) each word occured in the trial. 'num_words_sess' is how many epochs were in each session. """

    print('GETTING RECALLS')
    # RECALLS AND WORD INFO
    recalls_sess = np.asarray(words.recalled)
    rec = np.where(recalls_sess == 1)[0] # indices of recalled words
    not_rec = np.where(recalls_sess == 0)[0] # indices of not recalled words
    # GET INFORMATION ON WORD LISTS AND TIMING
    word_lists = np.asarray(words.list)
    lists = np.unique(word_lists)
    num_lists = np.size(lists)
    num_words_sess = np.size(recalls_sess)
    return {'recalls':recalls_sess, 'rec': rec, 'not_rec': not_rec, 'lists': lists, 'num_lists':
            num_lists, 'word_lists': word_lists,'num_words_sess': num_words_sess}

def get_powers(ts, mb, buffer, f_min, f_max, num_freqs):

    """CALCAULTES POWER MATRIX, LOG TRANSFORMS, AND AVERAGES ACROSS TIME. 
       Inputs: 'ts' is TimeSeriesX data for the subject, 'mb' is either 'M' for monopolar or 'B' for bipolar, 'buffer' is buffer surrounding each word in seconds, f_min and f_max are the min and max frequencies, num_freqs is number of log-spaced bins.
        Output: 2-dimensional array of size $num_words X ($num_channels/pairs$ X num_freqs$). For example, if there are 8 frequency bins, the indices [2,range(8,16)] would be the 8 powers for the third epoch, and second electrode."""

    print('CALCULATING POWERS')
    # CREATE THE MORLET WAVELET FILTER AND FIND POWER OF TIMESERIES
    wf = MorletWaveletFilter(time_series=ts, freqs=np.logspace(np.log10(f_min), np.log10(f_max), num_freqs), output='power')
    pow_wavelet, phase_wavelet = wf.filter()
    pow_wavelet = pow_wavelet.remove_buffer(duration=buffer)
    np.log10(pow_wavelet.data, out=pow_wavelet.data) #Log transform powers

    #LABEL MONOPOLAR AND BIPOLAR DIFFERENTLY
    if mb == 'M':
        pow_wavelet = pow_wavelet.transpose("events", "channels", "frequency", "time")
    else:
        pow_wavelet = pow_wavelet.transpose("events", "bipolar_pairs", "frequency", "time")
    #Average across time and reshape to two dimensions
    mean_powers_time_avg = pow_wavelet.mean(skipna=True, axis=-1)
    mean_powers_rs = mean_powers_time_avg.data.reshape(mean_powers_time_avg.data.shape[0], -1)
    return mean_powers_rs


#################### STD KUR ELECTRODE EPOCH REMOVAL; GET COEFFICIENTS; GET EXCLUSION METRICS###########################################
def std_kur_coefs_exclusion(sub, subpath, outpath, rand_iter=0):
    """ Automated statistical thresholding of electrodes/epochs based on standard deviation and kurtosis.

        Inputs: 'sub': string of subject ID, 'subpath': directory where subdata is, 'outpath': where you want data saved to, rand_iter: how many iterations of random removal you want

        Output: coefs: classifier coefficient matrix for each session (shape is num_sessions X num_lists X (8*num_channels))
                exclusions['car'/'bi']['eps'/'elecs']['std'/'kur'][$threshold$]: Indices of electrodes or epochs that were removed based on their standard deviation or kurtosis at a given threshold in common average or bipolar data
                output['mono'/'bi']['aucs'/'ts']['elec'/'epoch']['std'/'kur'][$threshold$]['rand'/'not_rand']: output metric (t-stat or AUC) after removing the given data for the given metric at a certain threshold. rand is for random removal, not_rand is for targeted removal."""
                
#for sub in ['R1001P']:
    #DEFINE CLASSIFIER AND INITIAL TERMS
    try:
        lr_classifier = LogisticRegression(C=7.2e-4, penalty='l2', class_weight='balanced', solver='liblinear')
        MB = 'MB' #remove the M or B if you want to remove the Monopolar or Bipolar analysis
        MU = 'MU' #remove the M or U if you want to remove the Multivariate or Univariate analysis
        CE = 'CE' #remove the C or E if you want to remove the Channel or Epoch exclusion analysis
        SK = 'SK' #remove the S or K if you want to remove the Standard deviation or Kurtosis analysis
        z_threshs = [0,.5,1,1.5,2,2.5,3,'none'] #z-thresholds for variance and kurtosis

        # CREATE OUTPUT DICTIONARIES
        coefs = {} #Where classifier coefficients are stored
        coefs['car'] = []
        coefs['pairs'] = []

        exclusions={} #Where indices of excluded epochs/electrodes are stored
        exclusions['car'] = {}
        exclusions['pairs'] = {}
        exclusions['car']['eps']={}
        exclusions['car']['elecs']={}
        exclusions['car']['recall_labels']=[]
        exclusions['pairs']['eps']={}
        exclusions['pairs']['elecs']={}
        exclusions['pairs']['recall_labels']=[]
        for mode_key in ['car', 'pairs']:
            for obj_key in ['eps','elecs']:
                for metr_key in ['std', 'kur']:
                    exclusions[mode_key][obj_key][metr_key] = {}
                    for thresh in z_threshs:
                        exclusions[mode_key][obj_key][metr_key][str(thresh)] = []

        output = {}            
        if 'M' in MB:
            output['car'] = {}
        if 'B' in MB:
            output['pairs'] = {}
        for key1 in list(output.keys()):
            if 'M' in MU:
                output[key1]['aucs'] = {}
            if 'U' in MU:
                output[key1]['ts'] = {}
            for key2 in list(output[key1].keys()):
                if 'C' in CE:
                    output[key1][key2]['elec'] = {}
                if 'E' in CE:
                    output[key1][key2]['epoch'] = {}
                for key3 in list(output[key1][key2].keys()):
                    if 'S' in SK:
                        output[key1][key2][key3]['std'] = {}
                    if 'K' in SK:
                        output[key1][key2][key3]['kur'] = {}
                    for key4 in list(output[key1][key2][key3].keys()):
                        for thresh in z_threshs:
                            output[key1][key2][key3][key4][str(thresh)] = {}
                            for rand_key in ['not_rand','rand']:
                                output[key1][key2][key3][key4][str(thresh)][rand_key] = []

        # IMPORT SUBJECT DATA
        data = np.load(subpath+'/'+sub+'_data.npz')['data'][()]
        events = data['events']
        exps = data['exps']
        sessions = data['sessions']
        channels = data['contacts']
        num_channels = data['num_contacts']
        pairs = data['pairs']
        num_pairs = data['num_pairs']
        recalls = data['recalls']
        rec = data['rec']
        not_rec = data['not_rec']
        lists = data['lists']
        num_lists = data['num_lists']
        word_lists = data['word_lists']
        num_words_sess = data['num_words_sess']
        powers_mono_uni = data['powers_car_uni']
        powers_pairs_uni = data['powers_pairs_uni']
        powers_mono_multi = data['powers_car_multi']
        powers_pairs_multi = data['powers_pairs_multi']
        stds_mono = data['stds_car']
        stds_pairs = data['stds_pairs']
        kurs_mono = data['kurs_car']
        kurs_pairs = data['kurs_pairs']

        for sess in range(np.size(sessions)): 
            # MAKE SURE THERE ARE WORDS RECALLED IN THE SESSION
            if np.size(np.unique(recalls[sess]))<2:
                np.savez('/scratch/stevenmeisler/Meis19_REVISIONS/std_kur_NEW_errors/'+sub+'_sess_'+str(sess)+'_error', error='NO RECALL')
                continue

            # MAKE SURE NO MISMATCH IN NUMBER OF WORDS
            if powers_mono_multi[sess].shape[0] != recalls[sess].size:
                np.savez('/scratch/stevenmeisler/Meis19_REVISIONS/std_kur_NEW_errors/'+sub+'_sess_'+str(sess)+'_error', error='MISMATCH')
                continue

            # CHECK FOR NaNs CHANNELS/PAIRS AND REMOVE THEM
            zscored_mono = zscore(powers_mono_multi[sess], axis=0, ddof=1) # ZSCORE POWERS ALONG EPOCH
            zscored_pairs = zscore(powers_pairs_multi[sess], axis=0, ddof=1) # ZSCORE POWERS ALONG EPOCH
            chan_nans = [int(x/8) for x in np.unique(np.where(np.isnan(zscored_mono)==True)[1]) if x % 8 == 0]
            pairs_nans = [int(x/8) for x in np.unique(np.where(np.isnan(zscored_pairs)==True)[1]) if x % 8 == 0]

            # REMOVE NaNs if they exist and update appropriate metrics
            if np.size(chan_nans)>0 or np.size(pairs_nans)>0:
                num_channels[sess]-=np.size(chan_nans)
                num_pairs[sess]-=np.size(pairs_nans)
                powers_mono_uni[sess] = np.reshape(powers_mono_uni[sess][~numpy.isnan(zscored_mono)],[num_words_sess[sess],-1])
                powers_mono_multi[sess] = np.reshape(powers_mono_multi[sess][~numpy.isnan(zscored_mono)],[num_words_sess[sess],-1])
                powers_pairs_uni[sess] = np.reshape(powers_pairs_uni[sess][~numpy.isnan(zscored_pairs)],[num_words_sess[sess],-1])
                powers_pairs_multi[sess] = np.reshape(powers_pairs_multi[sess][~numpy.isnan(zscored_pairs)],[num_words_sess[sess],-1])
                zscored_mono = np.reshape(zscored_mono[~numpy.isnan(zscored_mono)],[num_words_sess[sess],-1])
                zscored_pairs = np.reshape(zscored_pairs[~numpy.isnan(zscored_pairs)],[num_words_sess[sess],-1])
                stds_mono[sess] = np.delete(stds_mono[sess], chan_nans, axis=1)
                kurs_mono[sess] = np.delete(kurs_mono[sess], chan_nans, axis=1)
                stds_pairs[sess] = np.delete(stds_pairs[sess], pairs_nans, axis=1)
                kurs_pairs[sess] = np.delete(kurs_pairs[sess], pairs_nans, axis=1)
        
             # CALCULATE CLASSIFIER COEFFICIENTS
            coefs_mono = []
            coefs_pairs  = []
            recall_probs_mono = []
            recall_probs_pairs = []  
            for l_val in np.unique(word_lists[sess]): 
                l_train = list(np.unique(word_lists[sess])) #GET AND SEPERATE TRAINING AND VALIDATION DATA
                l_train.remove(l_val)
                val_mask=(word_lists[sess]==l_val)
                train_mask=(val_mask==False)
                rec_val=np.asarray(recalls[sess])[val_mask]
                rec_train=np.asarray(recalls[sess])[train_mask]
                #MONO
                zscored_val=zscored_mono[val_mask,:]
                zscored_train=zscored_mono[train_mask,:]
                lr_classifier.fit(zscored_train, rec_train) # FIT CLASSIFIER
                validation_recall_prob_array = lr_classifier.predict_proba(zscored_val)[:,1]
                recall_probs_mono.extend(validation_recall_prob_array)
                coef = lr_classifier.coef_[0] # GET CLASSIFIER WEIGHTS
                coefs_mono.append(coef)
                #BI
                zscored_val=zscored_pairs[val_mask,:]
                zscored_train=zscored_pairs[train_mask,:]
                lr_classifier.fit(zscored_train, rec_train) # FIT CLASSIFIER
                validation_recall_prob_array = lr_classifier.predict_proba(zscored_val)[:,1]
                recall_probs_pairs.extend(validation_recall_prob_array)
                coef = lr_classifier.coef_[0] # GET CLASSIFIER WEIGHTS
                coefs_pairs.append(coef)
            coefs['car'].append(coefs_mono)
            coefs['pairs'].append(coefs_pairs)
            recall_probs_rs = np.reshape(recall_probs_mono, np.size(recall_probs_mono)) #RESHAPE FOR ROC FUNCTION TO ALLOW
            exclusions['car']['recall_labels'].append(recall_probs_rs)
            recall_probs_rs = np.reshape(recall_probs_pairs, np.size(recall_probs_pairs)) #RESHAPE FOR ROC FUNCTION TO ALLOW
            exclusions['pairs']['recall_labels'].append(recall_probs_rs)

            # Get exclusion metrics
            # Z-score the average standard deviation or kurtosis values along epochs or electrodes
            # *_elecs means average across electrodes before z-scoring, *_eps means average across epochs before z-scoring
            z_stds_mono_elecs = zscore(np.mean(stds_mono[sess],axis=0))
            z_stds_mono_eps = zscore(np.mean(stds_mono[sess],axis=1))
            z_kurs_mono_elecs = zscore(np.mean(kurs_mono[sess],axis=0))
            z_kurs_mono_eps = zscore(np.mean(kurs_mono[sess],axis=1))
            z_stds_pairs_elecs = zscore(np.mean(stds_pairs[sess],axis=0))
            z_stds_pairs_eps = zscore(np.mean(stds_pairs[sess],axis=1))
            z_kurs_pairs_elecs = zscore(np.mean(kurs_pairs[sess],axis=0))
            z_kurs_pairs_eps = zscore(np.mean(kurs_pairs[sess],axis=1))

            for thresh in z_threshs[:-1]:
                excluded = np.where(z_stds_mono_elecs > thresh)[0]
                exclusions['car']['elecs']['std'][str(thresh)].append(excluded)
                excluded = np.where(z_stds_mono_eps > thresh)[0]
                exclusions['car']['eps']['std'][str(thresh)].append(excluded)
                excluded = np.where(z_kurs_mono_elecs > thresh)[0]
                exclusions['car']['elecs']['kur'][str(thresh)].append(excluded)
                excluded = np.where(z_kurs_mono_eps > thresh)[0]
                exclusions['car']['eps']['kur'][str(thresh)].append(excluded)
                excluded = np.where(z_stds_pairs_elecs > thresh)[0]
                exclusions['pairs']['elecs']['std'][str(thresh)].append(excluded)
                excluded = np.where(z_stds_pairs_eps > thresh)[0]
                exclusions['pairs']['eps']['std'][str(thresh)].append(excluded)
                excluded = np.where(z_kurs_pairs_elecs > thresh)[0]
                exclusions['pairs']['elecs']['kur'][str(thresh)].append(excluded)
                excluded = np.where(z_kurs_pairs_eps > thresh)[0]
                exclusions['pairs']['eps']['kur'][str(thresh)].append(excluded)
            
            #RUN CLASSIFIER ANALYSES AND ADD RESULTS TO OUTPUT DICTIONARY
            if 'M' in MU: #MULTIVARIATE
                if 'C' in CE: #REMOVE CHANNELS
                    if 'S' in SK: #STANDARD DEVIATION
                        auc, auc_rand = multivar_elec(powers_mono_multi[sess], z_stds_mono_elecs, num_channels[sess],
                                                      z_threshs, recalls[sess], word_lists[sess], rand_iteration = rand_iter)
                        for ind in range(np.size(auc)):
                            key5 = list(output[key1][key2][key3][key4].keys())[ind]
                            output['car']['aucs']['elec']['std'][key5]['not_rand'].append(auc[ind])
                            output['car']['aucs']['elec']['std'][key5]['rand'].append(auc_rand[ind])

                    if 'K' in SK: #KURTOSIS
                        auc, auc_rand = multivar_elec(powers_mono_multi[sess], z_kurs_mono_elecs, num_channels[sess],
                                                      z_threshs, recalls[sess], word_lists[sess], rand_iteration = rand_iter)
                        for ind in range(np.size(auc)):
                            key5 = list(output[key1][key2][key3][key4].keys())[ind]
                            output['car']['aucs']['elec']['kur'][key5]['not_rand'].append(auc[ind])
                            output['car']['aucs']['elec']['kur'][key5]['rand'].append(auc_rand[ind])

                if 'E' in CE: #REMOVE EPOCHS
                    if 'S' in SK: #STANDARD DEVIATION
                        auc, auc_rand = multivar_ep(powers_mono_multi[sess], z_stds_mono_eps, num_channels[sess], z_threshs,
                                                    recalls[sess], word_lists[sess], rand_iteration = rand_iter)
                        for ind in range(np.size(auc)):
                            key5 = list(output[key1][key2][key3][key4].keys())[ind]
                            output['car']['aucs']['epoch']['std'][key5]['not_rand'].append(auc[ind])
                            output['car']['aucs']['epoch']['std'][key5]['rand'].append(auc_rand[ind])

                    if 'K' in SK: #KURTOSIS
                        auc, auc_rand = multivar_ep(powers_mono_multi[sess], z_kurs_mono_eps, num_channels[sess], z_threshs,
                                                    recalls[sess], word_lists[sess], rand_iteration = rand_iter)
                        for ind in range(np.size(auc)):
                            key5 = list(output[key1][key2][key3][key4].keys())[ind]
                            output['car']['aucs']['epoch']['kur'][key5]['not_rand'].append(auc[ind])
                            output['car']['aucs']['epoch']['kur'][key5]['rand'].append(auc_rand[ind])

            if 'U' in MU: #UNIVARIATE
                if 'C' in CE: #REMOVE CHANNELS
                    if 'S' in SK: #STANDARD DEVIATION
                        t, t_rand = univar_elec(powers_mono_uni[sess],z_stds_mono_elecs, num_channels[sess], z_threshs,
                                    recalls[sess], rand_iteration = rand_iter)
                        for ind in range(np.size(t)):
                            key5 = list(output[key1][key2][key3][key4].keys())[ind]
                            output['car']['ts']['elec']['std'][key5]['not_rand'].append(t[ind])
                            output['car']['ts']['elec']['std'][key5]['rand'].append(t_rand[ind])

                    if 'K' in SK: #KURTOSIS
                        t, t_rand = univar_elec(powers_mono_uni[sess], z_kurs_mono_elecs, num_channels[sess], z_threshs,
                                    recalls[sess], rand_iteration = rand_iter)
                        for ind in range(np.size(t)):
                            key5 = list(output[key1][key2][key3][key4].keys())[ind]
                            output['car']['ts']['elec']['kur'][key5]['not_rand'].append(t[ind])
                            output['car']['ts']['elec']['kur'][key5]['rand'].append(t_rand[ind])

                if 'E' in CE: #REMOVE EPOCHS
                    if 'S' in SK: #STANDARD DEVIATION
                        t, t_rand = univar_ep(powers_mono_uni[sess], z_stds_mono_eps, num_channels[sess], z_threshs,
                                    recalls[sess], rand_iteration = rand_iter)
                        for ind in range(np.size(t)):
                            key5 = list(output[key1][key2][key3][key4].keys())[ind]
                            output['car']['ts']['epoch']['std'][key5]['not_rand'].append(t[ind])
                            output['car']['ts']['epoch']['std'][key5]['rand'].append(t_rand[ind])

                    if 'K' in SK: #KURTOSIS
                        t, t_rand = univar_ep(powers_mono_uni[sess], z_kurs_mono_eps, num_channels[sess], z_threshs,
                                    recalls[sess], rand_iteration = rand_iter)
                        for ind in range(np.size(t)):
                            key5 = list(output[key1][key2][key3][key4].keys())[ind]
                            output['car']['ts']['epoch']['kur'][key5]['not_rand'].append(t[ind])
                            output['car']['ts']['epoch']['kur'][key5]['rand'].append(t_rand[ind])

            #BIPOLAR ANALYSIS
            if "B" in MB:
                if 'M' in MU: #MULTIVARIATE 
                    #RUN ANALYSES AND ADD RESULTS TO OUTPUT DICTIONARY
                    if 'C' in CE: #REMOVE CHANNELS
                        if 'S' in SK: #STANDARD DEVIATION
                            auc, auc_rand = multivar_elec(powers_pairs_multi[sess], z_stds_pairs_elecs, num_pairs[sess], 
                                         z_threshs, recalls[sess], word_lists[sess], rand_iteration = rand_iter)
                            for ind in range(np.size(auc)):
                                key5 = list(output[key1][key2][key3][key4].keys())[ind]
                                output['pairs']['aucs']['elec']['std'][key5]['not_rand'].append(auc[ind])
                                output['pairs']['aucs']['elec']['std'][key5]['rand'].append(auc_rand[ind])

                        if 'K' in SK: #KURTOSIS
                            auc, auc_rand = multivar_elec(powers_pairs_multi[sess], z_kurs_pairs_elecs, num_pairs[sess], 
                                         z_threshs, recalls[sess], word_lists[sess], rand_iteration = rand_iter)
                            for ind in range(np.size(auc)):
                                key5 = list(output[key1][key2][key3][key4].keys())[ind]
                                output['pairs']['aucs']['elec']['kur'][key5]['not_rand'].append(auc[ind])
                                output['pairs']['aucs']['elec']['kur'][key5]['rand'].append(auc_rand[ind])

                    if 'E' in CE: #REMOVE EPOCHS
                        if 'S' in SK: #STANDARD DEVIATION
                            auc, auc_rand = multivar_ep(powers_pairs_multi[sess], z_stds_pairs_eps, num_pairs[sess],
                                        z_threshs, recalls[sess], word_lists[sess], rand_iteration = rand_iter)
                            for ind in range(np.size(auc)):
                                key5 = list(output[key1][key2][key3][key4].keys())[ind]
                                output['pairs']['aucs']['epoch']['std'][key5]['not_rand'].append(auc[ind])
                                output['pairs']['aucs']['epoch']['std'][key5]['rand'].append(auc_rand[ind])

                        if 'K' in SK: #KURTOSIS
                            auc, auc_rand = multivar_ep(powers_pairs_multi[sess], z_kurs_pairs_eps, num_pairs[sess],
                                        z_threshs, recalls[sess], word_lists[sess], rand_iteration = rand_iter)
                            for ind in range(np.size(auc)):
                                key5 = list(output[key1][key2][key3][key4].keys())[ind]
                                output['pairs']['aucs']['epoch']['kur'][key5]['not_rand'].append(auc[ind])
                                output['pairs']['aucs']['epoch']['kur'][key5]['rand'].append(auc_rand[ind])

                if 'U' in MU: # UNIVARIATE
                    if 'C' in CE: #REMOVE CHANNELS
                        if 'S' in SK: #STANDARD DEVIATION
                            t, t_rand = univar_elec(powers_pairs_uni[sess], z_stds_pairs_elecs, num_pairs[sess],
                                        z_threshs, recalls[sess], rand_iteration = rand_iter)
                            for ind in range(np.size(t)):
                                key5 = list(output[key1][key2][key3][key4].keys())[ind]
                                output['pairs']['ts']['elec']['std'][key5]['not_rand'].append(t[ind])
                                output['pairs']['ts']['elec']['std'][key5]['rand'].append(t_rand[ind])

                        if 'K' in SK: #KURTOSIS
                            t, t_rand = univar_elec(powers_pairs_uni[sess], z_kurs_pairs_elecs, num_pairs[sess],
                                        z_threshs, recalls[sess], rand_iteration = rand_iter)
                            for ind in range(np.size(t)):
                                key5 = list(output[key1][key2][key3][key4].keys())[ind]
                                output['pairs']['ts']['elec']['kur'][key5]['not_rand'].append(t[ind])
                                output['pairs']['ts']['elec']['kur'][key5]['rand'].append(t_rand[ind])

                    if 'E' in CE: #REMOVE EPOCHS
                        if 'S' in SK: #STANDARD DEVIATION
                            t, t_rand = univar_ep(powers_pairs_uni[sess], z_stds_pairs_eps, num_pairs[sess],
                                        z_threshs, recalls[sess], rand_iteration = rand_iter)
                            for ind in range(np.size(t)):
                                key5 = list(output[key1][key2][key3][key4].keys())[ind]
                                output['pairs']['ts']['epoch']['std'][key5]['not_rand'].append(t[ind])
                                output['pairs']['ts']['epoch']['std'][key5]['rand'].append(t_rand[ind])

                        if 'K' in SK: #KURTOSIS
                            t, t_rand = univar_ep(powers_pairs_uni[sess], z_kurs_pairs_eps, num_pairs[sess],
                                        z_threshs, recalls[sess], rand_iteration = rand_iter)
                            for ind in range(np.size(t)):
                                key5 = list(output[key1][key2][key3][key4].keys())[ind]
                                output['pairs']['ts']['epoch']['kur'][key5]['not_rand'].append(t[ind])
                                output['pairs']['ts']['epoch']['kur'][key5]['rand'].append(t_rand[ind])

        # SAVE OUT DATA
        np.savez(outpath+'/'+sub+'_std_kur', output=output)
        np.savez(outpath+'/'+sub+'_coefs', output=coefs)
        np.savez(outpath+'/'+sub+'_exclusion_metrics', output=exclusions)

    except Exception as e:
        np.savez(outpath+'/'+sub+'_error', error=e)
#####################################################
def manual_elecs(sub, subpath, outpath, elec_path):
    """ Performs manual electrode removal based on neurologist identification.
    INPUTS: 'sub': String of subject name, 'subpath': directory where subject data is stored, 'elec_path': directory where electrode category annotations are stored.
'outpath': where to save outputs

    OUTPUTS: Dictionary with indices of removed data, and resulting AUCS and t-stats
    """

    lr_classifier = LogisticRegression(C=7.2e-4, penalty='l2', class_weight='balanced', solver='liblinear')
    buffer = 1.5
    #INITIALIZE OUTPUTS
    output_mono = {'inds_removed_sil':[], 'aucs_sil':[], 'ts_sil':[],
                    'inds_removed_s':[], 'aucs_s':[], 'ts_s':[],
                  'inds_removed_i':[], 'aucs_i':[], 'ts_i':[],
                  'inds_removed_l':[], 'aucs_l':[], 'ts_l':[],
                  'inds_removed_si':[], 'aucs_si':[], 'ts_si':[],
                  'inds_removed_il':[],  'aucs_il':[], 'ts_il':[],
                   'inds_removed_sl':[], 'aucs_sl':[], 'ts_sl':[],
                  'inds_removed_none':[], 'aucs_none':[], 'ts_none':[]}
    output_pairs = {'inds_removed_sil':[], 'aucs_sil':[], 'ts_sil':[],
                    'inds_removed_s':[], 'aucs_s':[], 'ts_s':[],
                  'inds_removed_i':[], 'aucs_i':[], 'ts_i':[],
                  'inds_removed_l':[], 'aucs_l':[], 'ts_l':[],
                  'inds_removed_si':[], 'aucs_si':[], 'ts_si':[],
                  'inds_removed_il':[],  'aucs_il':[], 'ts_il':[],
                   'inds_removed_sl':[], 'aucs_sl':[], 'ts_sl':[],
                  'inds_removed_none':[], 'aucs_none':[], 'ts_none':[]}
    
    try:
        # IMPORT SUBJECT DATA
        data = np.load(subpath+'/'+sub+'_data.npz')['data'][()]
        events = data['events']
        exps = data['exps']
        sessions = data['sessions']
        channels_sub = data['contacts']
        num_channels = data['num_contacts']
        pairs = data['pairs']
        num_pairs = data['num_pairs']
        recalls = data['recalls']
        rec = data['rec']
        not_rec = data['not_rec']
        lists = data['lists']
        num_lists = data['num_lists']
        word_lists = data['word_lists']
        #offsets = data['offsets']
        num_words_sess = data['num_words_sess']
        powers_mono_uni = data['powers_car_uni']
        powers_pairs_uni = data['powers_pairs_uni']
        powers_mono_multi = data['powers_car_multi']
        powers_pairs_multi = data['powers_pairs_multi']
        stds_mono = data['stds_car']
        stds_pairs = data['stds_pairs']
        kurs_mono = data['kurs_car']
        kurs_pairs = data['kurs_pairs']

        # LOAD ELECTRODE CATEGORIES FILE
        elec_cat_file = '/data10/RAM/subjects/' + sub + '/docs/electrode_categories.txt'
        f = open(elec_cat_file, 'r')
        seizure = []
        interictal = []
        lesion = []
        bad = []

        #APPEND ELECTRODE NAMES TO THEIR PROPER CATEGORY
        append_var = -3
        for line in f:
            if append_var == -2:
                bad.append(line)
            if 'ad' in line:
                append_var = -2
            if append_var == -1:
                lesion.append(line)
            if 'lesi' in line:
                append_var = -1
            if append_var == 0:
                interictal.append(line)
            if 'ictal' in line:
                append_var = 0
            if append_var == 1:
                seizure.append(line)
            if 'zure' in line:
                append_var = 1

        # STRIP UNNECESSARY TEXT AND LINE DELIMITERS
        while True:
            try:
                bad.remove('\n')
            except:
                break
        bads = [s.strip('\n') for s in bad]

        while True:
            try:
                seizure.remove('\n')
            except:
                break
        seizures = [s.strip('\n') for s in seizure]

        while True:
            try:
                interictal.remove('\n')
            except:
                break
        interictals = [s.strip('\n') for s in interictal]

        while True:
            try:
                lesion.remove('\n')
            except:
                break
        lesions = [s.strip('\n') for s in lesion]
        #for sess in [0]:
        for sess in range(np.size(sessions)):
            print(sess)
            # Indices given channels
            pairs_seizure = []
            pairs_interictal = []
            pairs_lesion = []
            chans_seizure = []
            chans_interictal = []
            chans_lesion = []
            pairs_bad = []
            chans_bad = []

            chans = np.asarray(channels_sub[0])[:,0]
            indx_count = 0
            for pair_count in range(np.asarray(pairs[sess]).shape[0]):
                pair_info = np.asarray(pairs[sess])[pair_count]
                labels=pair_info[2].split('-')
                channels_pair = np.asarray(pairs[sess])[pair_count,[0,1]]
                chan_count = 0
                for label in labels:
                    if label in seizures:
                        pairs_seizure.append(indx_count)
                        chans_seizure.append(np.where(chans==channels_pair[chan_count])[0])
                    if label in interictals:
                        pairs_interictal.append(indx_count)
                        chans_interictal.append(np.where(chans==channels_pair[chan_count])[0])
                    if label in lesions:
                        pairs_lesion.append(indx_count)
                        chans_lesion.append(np.where(chans==channels_pair[chan_count])[0])
                    if label in bads:
                        pairs_bad.append(indx_count)
                        chans_bad.append(np.where(chans==channels_pair[chan_count])[0])
                    chan_count += 1
                indx_count += 1

            pairs_seizure = np.unique(pairs_seizure)
            chans_seizure = np.unique(chans_seizure)
            pairs_interictal = np.unique(pairs_interictal)
            chans_interictal = np.unique(chans_interictal)
            pairs_lesion = np.unique(pairs_lesion)
            chans_lesion = np.unique(chans_lesion)
            pairs_bad = np.unique(pairs_bad)
            chans_bad = np.unique(chans_bad)
            zscored_mono = zscore(powers_mono_multi[0], axis=0, ddof=1) # ZSCORE POWERS ALONG EPOCH
            zscored_pairs = zscore(powers_pairs_multi[0], axis=0, ddof=1) # ZSCORE POWERS ALONG EPOCH
            chan_nans = [int(x/8) for x in np.unique(np.where(np.isnan(zscored_mono)==True)[1]) if x % 8 == 0]
            pairs_nans = [int(x/8) for x in np.unique(np.where(np.isnan(zscored_pairs)==True)[1]) if x % 8 == 0]

            #S
            inds_to_remove = np.sort(np.unique(np.concatenate((chans_seizure,chan_nans))))
            auc,t = manual_elec_classification(powers_mono_multi[sess], powers_mono_uni[sess], num_channels[sess], inds_to_remove,
                                recalls[sess], word_lists[sess])
            output_mono['aucs_s'].append(auc)
            output_mono['ts_s'].append(t)
            output_mono['inds_removed_s'].append(chans_seizure)
            inds_to_remove = np.sort(np.unique(np.concatenate((pairs_seizure, pairs_nans))))
            auc,t = manual_elec_classification(powers_pairs_multi[sess], powers_pairs_uni[sess], num_pairs[sess], inds_to_remove,
                                recalls[sess], word_lists[sess])
            output_pairs['aucs_s'].append(auc)
            output_pairs['ts_s'].append(t)
            output_pairs['inds_removed_s'].append(pairs_seizure)

            #I
            inds_to_remove = np.sort(np.unique(np.concatenate((chans_interictal,chan_nans))))
            auc,t = manual_elec_classification(powers_mono_multi[sess], powers_mono_uni[sess], num_channels[sess], inds_to_remove,
                                recalls[sess], word_lists[sess])
            output_mono['aucs_i'].append(auc)
            output_mono['ts_i'].append(t)
            output_mono['inds_removed_i'].append(chans_interictal)
            inds_to_remove = np.sort(np.unique(np.concatenate((pairs_interictal, pairs_nans))))
            auc,t = manual_elec_classification(powers_pairs_multi[sess], powers_pairs_uni[sess], num_pairs[sess], inds_to_remove,
                                recalls[sess], word_lists[sess])
            output_pairs['aucs_i'].append(auc)
            output_pairs['ts_i'].append(t)
            output_pairs['inds_removed_i'].append(pairs_interictal)

            #L
            inds_to_remove = np.sort(np.unique(np.concatenate((chans_lesion,chan_nans))))
            auc,t = manual_elec_classification(powers_mono_multi[sess], powers_mono_uni[sess], num_channels[sess], inds_to_remove,
                                recalls[sess], word_lists[sess])
            output_mono['aucs_l'].append(auc)
            output_mono['ts_l'].append(t)
            output_mono['inds_removed_l'].append(chans_lesion)
            inds_to_remove = np.sort(np.unique(np.concatenate((pairs_lesion, pairs_nans))))
            auc,t = manual_elec_classification(powers_pairs_multi[sess], powers_pairs_uni[sess], num_pairs[sess], inds_to_remove,
                                recalls[sess], word_lists[sess])
            output_pairs['aucs_l'].append(auc)
            output_pairs['ts_l'].append(t)
            output_pairs['inds_removed_l'].append(pairs_lesion)

            #SI
            chans_si = np.sort(np.unique(np.concatenate((chans_seizure, chans_interictal))))
            inds_to_remove = np.sort(np.unique(np.concatenate((chans_si,chan_nans))))
            auc,t = manual_elec_classification(powers_mono_multi[sess], powers_mono_uni[sess], num_channels[sess], inds_to_remove,
                                recalls[sess], word_lists[sess])
            output_mono['aucs_si'].append(auc)
            output_mono['ts_si'].append(t)
            output_mono['inds_removed_si'].append(chans_si)
            pairs_si = np.sort(np.unique(np.concatenate((pairs_seizure, pairs_interictal))))
            inds_to_remove = np.sort(np.unique(np.concatenate((pairs_si, pairs_nans))))
            auc,t = manual_elec_classification(powers_pairs_multi[sess], powers_pairs_uni[sess], num_pairs[sess], inds_to_remove,
                                recalls[sess], word_lists[sess])
            output_pairs['aucs_si'].append(auc)
            output_pairs['ts_si'].append(t)
            output_pairs['inds_removed_si'].append(pairs_si)

            #SL
            chans_sl = np.sort(np.unique(np.concatenate((chans_seizure, chans_lesion))))
            inds_to_remove = np.sort(np.unique(np.concatenate((chans_sl,chan_nans))))
            auc,t = manual_elec_classification(powers_mono_multi[sess], powers_mono_uni[sess], num_channels[sess], inds_to_remove,
                                recalls[sess], word_lists[sess])
            output_mono['aucs_sl'].append(auc)
            output_mono['ts_sl'].append(t)
            output_mono['inds_removed_sl'].append(chans_sl)
            pairs_sl = np.sort(np.unique(np.concatenate((pairs_seizure, pairs_lesion))))
            inds_to_remove = np.sort(np.unique(np.concatenate((pairs_sl, pairs_nans))))
            auc,t = manual_elec_classification(powers_pairs_multi[sess], powers_pairs_uni[sess], num_pairs[sess], inds_to_remove,
                                recalls[sess], word_lists[sess])
            output_pairs['aucs_sl'].append(auc)
            output_pairs['ts_sl'].append(t)
            output_pairs['inds_removed_sl'].append(pairs_sl)

            #IL
            chans_il = np.sort(np.unique(np.concatenate((chans_interictal, chans_lesion))))
            inds_to_remove = np.sort(np.unique(np.concatenate((chans_il,chan_nans))))
            auc,t = manual_elec_classification(powers_mono_multi[sess], powers_mono_uni[sess], num_channels[sess], inds_to_remove,
                                recalls[sess], word_lists[sess])
            output_mono['aucs_il'].append(auc)
            output_mono['ts_il'].append(t)
            output_mono['inds_removed_il'].append(chans_il)
            pairs_il = np.sort(np.unique(np.concatenate((pairs_interictal, pairs_lesion))))
            inds_to_remove = np.sort(np.unique(np.concatenate((pairs_il, pairs_nans))))
            auc,t = manual_elec_classification(powers_pairs_multi[sess], powers_pairs_uni[sess], num_pairs[sess], inds_to_remove,
                                recalls[sess], word_lists[sess])
            output_pairs['aucs_il'].append(auc)
            output_pairs['ts_il'].append(t)
            output_pairs['inds_removed_il'].append(pairs_il)

            #SIL
            chans_sil = np.sort(np.unique(np.concatenate((chans_seizure, chans_interictal, chans_lesion))))
            inds_to_remove = np.sort(np.unique(np.concatenate((chans_sil,chan_nans))))
            auc,t = manual_elec_classification(powers_mono_multi[sess], powers_mono_uni[sess], num_channels[sess], inds_to_remove,
                                recalls[sess], word_lists[sess])
            output_mono['aucs_sil'].append(auc)
            output_mono['ts_sil'].append(t)
            output_mono['inds_removed_sil'].append(chans_sil)
            pairs_sil = np.sort(np.unique(np.concatenate((pairs_seizure, pairs_interictal, pairs_lesion))))
            inds_to_remove = np.sort(np.unique(np.concatenate((pairs_sil, pairs_nans))))
            auc,t = manual_elec_classification(powers_pairs_multi[sess], powers_pairs_uni[sess], num_pairs[sess], inds_to_remove,
                                recalls[sess], word_lists[sess])
            output_pairs['aucs_sil'].append(auc)
            output_pairs['ts_sil'].append(t)
            output_pairs['inds_removed_sil'].append(pairs_sil)

            #NONE
            inds_to_remove=chan_nans.copy()
            auc,t = manual_elec_classification(powers_mono_multi[sess], powers_mono_uni[sess], num_channels[sess], inds_to_remove,
                                recalls[sess], word_lists[sess])
            output_mono['aucs_none'].append(auc)
            output_mono['ts_none'].append(t)
            output_mono['inds_removed_none'].append(inds_to_remove)
            inds_to_remove=pairs_nans.copy()
            auc,t = manual_elec_classification(powers_pairs_multi[sess], powers_pairs_uni[sess], num_pairs[sess], inds_to_remove,
                                recalls[sess], word_lists[sess])
            output_pairs['aucs_none'].append(auc)
            output_pairs['ts_none'].append(t)
            output_pairs['inds_removed_none'].append(inds_to_remove)
        np.savez(outpath+'/'+sub+'_manual_elec', output={'car':output_mono, 'pairs': output_pairs})
    except Exception as e:
        print(e)
        np.savez(outpath+'/'+sub+'_ERROR', output=e)

def hfo_epochs(sub, subpath, outpath)
    """ Performs manual electrode removal based on neurologist identification.
    INPUTS: 'sub': String of subject name, 'subpath': directory where subject data is stored, 'outpath': where to save outputs

    OUTPUTS: Dictionary with classificaiton results for common_average and 
    """
    try:
        #Define output
        output={'car':{}, 'pairs':{}, 'hfo_metrics': {}}
        for key in ['car', 'pairs']:
            output[key]['aucs']=[]
            output[key]['ts']=[]
        output['hfo_metrics']['num_chans']=[]
        output['hfo_metrics']['num_pairs']=[]
        output['hfo_metrics']['ones']=[]
        output['hfo_metrics']['twos']=[]
        output['hfo_metrics']['fours']=[]
        output['hfo_metrics']['hfo_epochs']=[]
        output['hfo_metrics']['kept_epochs']=[]

        mat_path = '/scratch/liliana.camarillo/HFO_annotations_fr/all_files'
        epochs_annotated = []
        
        #Load subject data
        data = np.load(subpath+'/'+sub+'_data.npz')['data'][()]
        events = data['events']
        exps = data['exps']
        FR1_inds = [x for x in range(np.size(exps)) if exps[x]=='FR1']
        catFR1_inds = [x for x in range(np.size(exps)) if exps[x]=='catFR1']
        sessions = data['sessions']
        channels_sub = data['contacts']
        num_channels = data['num_contacts']
        pairs = data['pairs']
        num_pairs = data['num_pairs']
        recalls = data['recalls']
        rec = data['rec']
        not_rec = data['not_rec']
        lists = data['lists']
        num_lists = data['num_lists']
        word_lists = data['word_lists']
        num_words_sess = data['num_words_sess']
        powers_mono_uni = data['powers_car_uni']
        powers_pairs_uni = data['powers_pairs_uni']
        powers_mono_multi = data['powers_car_multi']
        powers_pairs_multi = data['powers_pairs_multi']

        for exp in ['FR1', 'CatFR1']:
            try: dat_HFO = sio.loadmat(mat_path+'/test'+sub+exp+'.mat')
            except: 
                continue
            if exp == 'CatFR1': exp='catFR1'
            exp_sess = []
            for sess in range(np.size(sessions)):
                if exps[sess] == exp: exp_sess.append(sess)
            annots = dat_HFO['HFO']
            hfo_pairs = dat_HFO['Channels']
            hfo_pairs_contacts = [list(hfo_pairs[0][x][1][0]) for x in range(hfo_pairs.shape[1])]
            ind_pairs_sess=[]
            pair_ind=0
            for [c1,c2] in zip(np.asarray([pairs[exp_sess[0]]['contact_1']][0]),np.asarray([pairs[exp_sess[0]]['contact_2']][0])):
                for cont in hfo_pairs_contacts:
                    if [c1,c2]==cont:
                        ind_pairs_sess.append(pair_ind)
                pair_ind+=1

            ind_pairs_freq_sess = []
            for ind in ind_pairs_sess:
                ind_pairs_freq_sess.extend(range(ind*8,ind*8+8))

            ind_mono_sess=[]
            for cont in range(np.size(np.unique(hfo_pairs_contacts))):
                if np.unique(hfo_pairs_contacts)[cont] in np.asarray(channels_sub[exp_sess[0]]['contact']):
                    ind_mono_sess.append(cont)

            ind_mono_freq_sess = []
            for ind in ind_mono_sess:
                ind_mono_freq_sess.extend(range(ind*8,ind*8+8))

            sess_no = 0
            for sess in exp_sess:
                num_pairs_sess = np.size(ind_pairs_sess)
                num_chans_sess = np.size(ind_mono_sess)
                word_cut_offs = [0]
                word_cut_offs += list(np.cumsum(np.asarray(num_words_sess)[exp_sess]))
                annots_ind_sess = range(word_cut_offs[sess_no], word_cut_offs[sess_no+1])
                annots_sess = annots[annots_ind_sess,:]
                output['hfo_metrics']['ones'].append(np.unique(np.where(annots_sess == 1)[0]))
                output['hfo_metrics']['twos'].append(np.unique(np.where(annots_sess == 2)[0]))
                output['hfo_metrics']['fours'].append(np.unique(np.where(annots_sess == 4)[0]))
                epochs_HFO = np.unique(list(np.unique(np.where(annots_sess == 1)[0])) + 
                      list(np.unique(np.where(annots_sess == 2)[0])) + list(np.unique(np.where(annots_sess == 4)[0])))
                epochs_kept = list(range(num_words_sess[sess]))
                for ep in epochs_HFO:
                    epochs_kept.remove(ep)
                output['hfo_metrics']['hfo_epochs'].append(epochs_HFO)
                output['hfo_metrics']['kept_epochs'].append(epochs_kept)

                #MAKE SURE THERE ARE RECALLED WORDS
                if np.size(np.unique(recalls[sess]))<2:
                    continue

                #MAKE SURE NO MISMATCH IN NUMBER OF WORDS
                if powers_mono_multi[sess].shape[0] != recalls[sess].size:
                    continue

                # LIMIT ONLY TO HFO ELECS
                powers_mono_multi[sess] = powers_mono_multi[sess][:,ind_mono_freq_sess]
                powers_mono_uni[sess] = powers_mono_uni[sess][:,ind_mono_freq_sess]
                powers_pairs_multi[sess] = powers_pairs_multi[sess][:,ind_pairs_freq_sess]
                powers_pairs_uni[sess] = powers_pairs_uni[sess][:,ind_pairs_freq_sess]
                # CHECK FOR NaNs CHANNELS/PAIRS AND REMOVE THEM
                zscored_mono = zscore(powers_mono_multi[sess], axis=0, ddof=1) # ZSCORE POWERS ALONG EPOCH
                zscored_pairs = zscore(powers_pairs_multi[sess], axis=0, ddof=1) # ZSCORE POWERS ALONG EPOCH
                chan_nans = [int(x/8) for x in np.unique(np.where(np.isnan(zscored_mono)==True)[1]) if x % 8 == 0]
                pairs_nans = [int(x/8) for x in np.unique(np.where(np.isnan(zscored_pairs)==True)[1]) if x % 8 == 0]

                # REMOVE NaNs if they exist and update appropriate metrics
                if np.size(chan_nans)>0 or np.size(pairs_nans)>0:
                    num_chans_sess-=np.size(chan_nans)
                    num_pairs_sess-=np.size(pairs_nans)
                    powers_mono_uni[sess] = np.reshape(powers_mono_uni[sess][~numpy.isnan(zscored_mono)],[num_words_sess[sess],-1])
                    powers_mono_multi[sess] = np.reshape(powers_mono_multi[sess][~numpy.isnan(zscored_mono)],[num_words_sess[sess],-1])
                    powers_pairs_uni[sess] = np.reshape(powers_pairs_uni[sess][~numpy.isnan(zscored_pairs)],[num_words_sess[sess],-1])
                    powers_pairs_multi[sess] = np.reshape(powers_pairs_multi[sess][~numpy.isnan(zscored_pairs)],[num_words_sess[sess],-1])

                recalls_sess = recalls[sess][epochs_kept]
                rec = np.where(recalls_sess==1)[0]
                not_rec = np.where(recalls_sess==0)[0]
                word_lists_sess = word_lists[sess][epochs_kept]
                #MULTIVARIATE_PAIRS
                powers_kept = powers_pairs_multi[sess][epochs_kept,:]
                zscored = zscore(powers_kept, axis=0, ddof=1) # ZSCORE POWERS ALONG EPOCH
                recall_probs = []
                for l_val in np.unique(word_lists_sess): 
                    l_train = list(np.unique(word_lists_sess)) #GET TRAINING AND VALIDATION DATA
                    l_train.remove(l_val)
                    val_mask=(word_lists_sess==l_val)
                    train_mask=(val_mask==False)
                    rec_val=np.asarray(recalls_sess)[val_mask]
                    rec_train=np.asarray(recalls_sess)[train_mask]
                    zscored_val=zscored[val_mask,:]
                    zscored_train=zscored[train_mask,:]
                    lr_classifier.fit(zscored_train, rec_train) # FIT AND EVALUTATE CLASSIFIER
                    validation_recall_prob_array = lr_classifier.predict_proba(zscored_val)[:,1]
                    recall_probs.extend(validation_recall_prob_array)
                recall_probs_rs = np.reshape(recall_probs, np.size(recall_probs)) #RESHAPE FOR ROC FUNCTION TO ALLOW
                auc=roc_auc_score(recalls_sess, recall_probs_rs) # CALCULATE ROC
                output['pairs']['aucs'].append(auc)

                #UNIVARIATE PAIRS
                t_stats=[]
                powers_kept = powers_pairs_uni[sess][epochs_kept,:]
                zscored = zscore(powers_kept, axis=0, ddof=1) # ZSCORE POWERS ALONG EPOCH
                zscore_mean_powers = np.mean(zscored.reshape(-1,8),axis=1).reshape(np.size(recalls_sess),num_pairs_sess)
                for elec in range(np.size(num_pairs_sess)):
                    powers_rec = zscore_mean_powers[rec, elec]
                    powers_rec_rs = np.reshape(powers_rec, np.size(powers_rec))
                    powers_not_rec = zscore_mean_powers[not_rec, elec]
                    powers_not_rec_rs = np.reshape(powers_not_rec, np.size(powers_not_rec))
                    t_stats.append(scipy.stats.ttest_ind(powers_rec_rs, powers_not_rec_rs, equal_var=False).statistic)
                output['pairs']['ts'].append(np.mean(t_stats))
                #ts.append(np.mean(t_stats))

                #MULTIVARIATE_MONO
                powers_kept = powers_mono_multi[sess][epochs_kept,:]
                zscored = zscore(powers_kept, axis=0, ddof=1) # ZSCORE POWERS ALONG EPOCH
                recall_probs = []
                for l_val in np.unique(word_lists_sess): 
                    l_train = list(np.unique(word_lists_sess)) #GET TRAINING AND VALIDATION DATA
                    l_train.remove(l_val)
                    val_mask=(word_lists_sess==l_val)
                    train_mask=(val_mask==False)
                    rec_val=np.asarray(recalls_sess)[val_mask]
                    rec_train=np.asarray(recalls_sess)[train_mask]
                    zscored_val=zscored[val_mask,:]
                    zscored_train=zscored[train_mask,:]
                    lr_classifier.fit(zscored_train, rec_train) # FIT AND EVALUTATE CLASSIFIER
                    validation_recall_prob_array = lr_classifier.predict_proba(zscored_val)[:,1]
                    recall_probs.extend(validation_recall_prob_array)
                recall_probs_rs = np.reshape(recall_probs, np.size(recall_probs)) #RESHAPE FOR ROC FUNCTION TO ALLOW
                auc=roc_auc_score(recalls_sess, recall_probs_rs) # CALCULATE ROC
                output['car']['aucs'].append(auc)
                #UNIVARIATE_MONO

                t_stats=[]
                powers_kept = powers_mono_uni[sess][epochs_kept,:]
                zscored = zscore(powers_kept, axis=0, ddof=1) # ZSCORE POWERS ALONG EPOCH
                zscore_mean_powers = np.mean(zscored.reshape(-1,8),axis=1).reshape(np.size(recalls_sess),num_chans_sess)
                for elec in range(np.size(num_chans_sess)):
                    powers_rec = zscore_mean_powers[rec, elec]
                    powers_rec_rs = np.reshape(powers_rec, np.size(powers_rec))
                    powers_not_rec = zscore_mean_powers[not_rec, elec]
                    powers_not_rec_rs = np.reshape(powers_not_rec, np.size(powers_not_rec))
                    t_stats.append(scipy.stats.ttest_ind(powers_rec_rs, powers_not_rec_rs, equal_var=False).statistic)
                output['car']['ts'].append(np.mean(t_stats))

                output['hfo_metrics']['num_pairs'].append(num_pairs_sess)
                output['hfo_metrics']['num_chans'].append(num_chans_sess)

                sess_no+=1

        np.savez(outpath+'/'+sub+'_hfo', outputs=output)  
    except Exception as e:
        np.savez(outpath+'/'+sub+'_hfo_error', outputs=e) 
        

## SUBSAMPLING ##
def std_kur_subsample(sub,subpath,outpath, iterations=100):
    """ Automated statistical thresholding of electrodes/epochs based on standard deviation and kurtosis.

        Inputs: sub: string of subject ID, subpath: where subject data is located, outpath: directory to save output to

        Output:  Dictionary indexable in the following way:
        ['car'/'pairs']['aucs'/'ts'][$percent_as_string$]['elec'/'epoch' (what is subsampled)]['elec'/'epoch' (what is thresholded)]
        ['std'/'kur'][$z_threshold as a string$].
"""
    #DEFINE CLASSIFIER AND INITIAL TERMS
    try:
        lr_classifier = LogisticRegression(C=7.2e-4, penalty='l2', class_weight='balanced', solver='liblinear')
        ss_iter = iterations
        rand_iter = None
        SS = ['20','40','60','80'] #deciles for subsampling, remove all but 100 if you don't want subsampling
        z_threshs = [2,2.5,3,'none'] #z-thresholds for variance and kurtosis


        output = {}            
        output['car'] = {}
        output['pairs'] = {}
        for key1 in list(output.keys()):
            output[key1]['aucs'] = {}
            output[key1]['ts'] = {}
            for key2 in list(output[key1].keys()):
                for ss in SS:
                    output[key1][key2][ss]={}
                    output[key1][key2][ss]['elec'] = {}
                    output[key1][key2][ss]['epoch'] = {}
                    output[key1][key2][ss]['elec']['elec'] = {}
                    output[key1][key2][ss]['elec']['epoch'] = {}
                    output[key1][key2][ss]['epoch']['elec'] = {}
                    output[key1][key2][ss]['epoch']['epoch'] = {}
                    for key3 in list(output[key1][key2][ss]['epoch'].keys()):
                        output[key1][key2][ss]['epoch'][key3]['std']={}
                        output[key1][key2][ss]['epoch'][key3]['kur']={}
                        output[key1][key2][ss]['elec'][key3]['std']={}
                        output[key1][key2][ss]['elec'][key3]['kur']={}
                        for key4 in list(output[key1][key2][ss]['epoch'][key3].keys()):
                            for thresh in z_threshs:
                                output[key1][key2][ss]['epoch'][key3][key4][str(thresh)] = []
                                output[key1][key2][ss]['elec'][key3][key4][str(thresh)] = []
        # IMPORT SUBJECT DATA
        data = np.load(subpath+'/'+sub+'_data.npz')['data'][()]
        events = data['events']
        exps = data['exps']
        sessions = data['sessions']
        channels = data['contacts']
        num_channels = data['num_contacts']
        pairs = data['pairs']
        num_pairs = data['num_pairs']
        recalls = data['recalls']
        rec = data['rec']
        not_rec = data['not_rec']
        lists = data['lists']
        num_lists = data['num_lists']
        word_lists = data['word_lists']
        num_words_sess = data['num_words_sess']
        powers_mono_uni = data['powers_car_uni']
        powers_pairs_uni = data['powers_pairs_uni']
        powers_mono_multi = data['powers_car_multi']
        powers_pairs_multi = data['powers_pairs_multi']
        stds_mono = data['stds_car']
        stds_pairs = data['stds_pairs']
        kurs_mono = data['kurs_car']
        kurs_pairs = data['kurs_pairs']

        for sess in range(np.size(sessions)): 
        # MAKE SURE THERE ARE WORDS RECALLED IN THE SESSION
            if np.size(np.unique(recalls[sess]))<2:
                continue

            # MAKE SURE NO MISMATCH IN NUMBER OF WORDS
            if powers_mono_multi[sess].shape[0] != recalls[sess].size:
                continue
            # CHECK FOR NaNs CHANNELS/PAIRS AND REMOVE THEM
            zscored_mono = zscore(powers_mono_multi[sess], axis=0, ddof=1) # ZSCORE POWERS ALONG EPOCH
            zscored_pairs = zscore(powers_pairs_multi[sess], axis=0, ddof=1) # ZSCORE POWERS ALONG EPOCH
            chan_nans = [int(x/8) for x in np.unique(np.where(np.isnan(zscored_mono)==True)[1]) if x % 8 == 0]
            pairs_nans = [int(x/8) for x in np.unique(np.where(np.isnan(zscored_pairs)==True)[1]) if x % 8 == 0]

            # REMOVE NaNs if they exist and update appropriate metrics
            if np.size(chan_nans)>0 or np.size(pairs_nans)>0:
                num_channels[sess]-=np.size(chan_nans)
                num_pairs[sess]-=np.size(pairs_nans)
                powers_mono_uni[sess] = np.reshape(powers_mono_uni[sess][~numpy.isnan(zscored_mono)],[num_words_sess[sess],-1])
                powers_mono_multi[sess] = np.reshape(powers_mono_multi[sess][~numpy.isnan(zscored_mono)],[num_words_sess[sess],-1])
                powers_pairs_uni[sess] = np.reshape(powers_pairs_uni[sess][~numpy.isnan(zscored_pairs)],[num_words_sess[sess],-1])
                powers_pairs_multi[sess] = np.reshape(powers_pairs_multi[sess][~numpy.isnan(zscored_pairs)],[num_words_sess[sess],-1])
                zscored_mono = np.reshape(zscored_mono[~numpy.isnan(zscored_mono)],[num_words_sess[sess],-1])
                zscored_pairs = np.reshape(zscored_pairs[~numpy.isnan(zscored_pairs)],[num_words_sess[sess],-1])
                stds_mono[sess] = np.delete(stds_mono[sess], chan_nans, axis=1)
                kurs_mono[sess] = np.delete(kurs_mono[sess], chan_nans, axis=1)
                stds_pairs[sess] = np.delete(stds_pairs[sess], pairs_nans, axis=1)
                kurs_pairs[sess] = np.delete(kurs_pairs[sess], pairs_nans, axis=1)

            for percent in [int(x) for x in SS]:
                for iteration in range(ss_iter):
                    chans_inds_iter = np.sort(random.sample(range(num_channels[sess]),int(num_channels[sess]*percent/100)))
                    pairs_inds_iter = np.sort(random.sample(range(num_pairs[sess]),int(num_pairs[sess]*percent/100)))
                    list_inds_iter = np.sort(random.sample([x+1 for x in range(num_lists[sess])],int(num_lists[sess]*percent/100)))
                    epoch_inds_iter = [ind for ind in range(np.size(word_lists[sess])) if word_lists[sess][ind] in list_inds_iter]

                    powers_mono_multi_elec = np.asarray(powers_mono_multi[sess])[:, np.concatenate([np.arange(ind*8, (ind+1)*8) 
                        for ind in chans_inds_iter])]
                    powers_pairs_multi_elec = np.asarray(powers_pairs_multi[sess])[:, np.concatenate([np.arange(ind*8, (ind+1)*8) 
                        for ind in pairs_inds_iter])]
                    powers_mono_uni_elec = np.asarray(powers_mono_uni[sess])[:, np.concatenate([np.arange(ind*8, (ind+1)*8) 
                        for ind in chans_inds_iter])]
                    powers_pairs_uni_elec = np.asarray(powers_pairs_uni[sess])[:, np.concatenate([np.arange(ind*8, (ind+1)*8) 
                        for ind in pairs_inds_iter])]

                    powers_mono_multi_ep = powers_mono_multi[sess][epoch_inds_iter]
                    powers_pairs_multi_ep = powers_pairs_multi[sess][epoch_inds_iter]
                    powers_mono_uni_ep = powers_mono_multi[sess][epoch_inds_iter]
                    powers_pairs_uni_ep = powers_pairs_multi[sess][epoch_inds_iter]

                    stds_mono_elec_iter = np.asarray(stds_mono[sess])[:,chans_inds_iter]
                    kurs_mono_elec_iter = np.asarray(kurs_mono[sess])[:,chans_inds_iter]
                    stds_pairs_elec_iter = np.asarray(stds_pairs[sess])[:,pairs_inds_iter]
                    kurs_pairs_elec_iter = np.asarray(kurs_pairs[sess])[:,pairs_inds_iter]
                    z_stds_mono_elec_elec = zscore(np.mean(stds_mono_elec_iter,axis=0))
                    z_stds_mono_elec_ep = zscore(np.mean(stds_mono_elec_iter,axis=1))
                    z_kurs_mono_elec_elec = zscore(np.mean(kurs_mono_elec_iter,axis=0))
                    z_kurs_mono_elec_ep = zscore(np.mean(kurs_mono_elec_iter,axis=1))
                    z_stds_pairs_elec_elec = zscore(np.mean(stds_pairs_elec_iter,axis=0))
                    z_stds_pairs_elec_ep = zscore(np.mean(stds_pairs_elec_iter,axis=1))
                    z_kurs_pairs_elec_elec = zscore(np.mean(kurs_pairs_elec_iter,axis=0))
                    z_kurs_pairs_elec_ep = zscore(np.mean(kurs_pairs_elec_iter,axis=1))

                    stds_mono_ep_iter = np.asarray(stds_mono[sess])[epoch_inds_iter,:]
                    kurs_mono_ep_iter = np.asarray(kurs_mono[sess])[epoch_inds_iter,:]
                    stds_pairs_ep_iter = np.asarray(stds_pairs[sess])[epoch_inds_iter,:]
                    kurs_pairs_ep_iter = np.asarray(kurs_pairs[sess])[epoch_inds_iter,:]
                    z_stds_mono_ep_elec = zscore(np.mean(stds_mono_ep_iter,axis=0))
                    z_stds_mono_ep_ep = zscore(np.mean(stds_mono_ep_iter,axis=1))
                    z_kurs_mono_ep_elec = zscore(np.mean(kurs_mono_ep_iter,axis=0))
                    z_kurs_mono_ep_ep = zscore(np.mean(kurs_mono_ep_iter,axis=1))
                    z_stds_pairs_ep_elec = zscore(np.mean(stds_pairs_ep_iter,axis=0))
                    z_stds_pairs_ep_ep = zscore(np.mean(stds_pairs_ep_iter,axis=1))
                    z_kurs_pairs_ep_elec = zscore(np.mean(kurs_pairs_ep_iter,axis=0))
                    z_kurs_pairs_ep_ep = zscore(np.mean(kurs_pairs_ep_iter,axis=1))

                    ##SUBSAMPLE ELECTRODES##
                    #MONO#

                    auc, auc_rand = multivar_elec(powers_mono_multi_elec, z_stds_mono_elec_elec, np.size(chans_inds_iter),
                                                          z_threshs, recalls[sess], word_lists[sess], rand_iteration = None)
                    for ind in range(np.size(auc)):
                        key5 = [str(z) for z in z_threshs][ind]
                        output['car']['aucs'][str(percent)]['elec']['elec']['std'][key5].append(auc[ind])



                    auc, auc_rand = multivar_elec(powers_mono_multi_elec, z_kurs_mono_elec_elec, np.size(chans_inds_iter),
                                                          z_threshs, recalls[sess], word_lists[sess], rand_iteration = None)
                    for ind in range(np.size(auc)):
                        key5 = [str(z) for z in z_threshs][ind]
                        output['car']['aucs'][str(percent)]['elec']['elec']['kur'][key5].append(auc[ind])



                    t, t_rand = univar_elec(powers_mono_uni_elec,z_stds_mono_elec_elec, np.size(chans_inds_iter), z_threshs,
                                        recalls[sess], rand_iteration = rand_iter)
                    for ind in range(np.size(t)):
                        key5 = [str(z) for z in z_threshs][ind]
                        output['car']['ts'][str(percent)]['elec']['elec']['std'][key5].append(t[ind])



                    t, t_rand = univar_elec(powers_mono_uni_elec,z_kurs_mono_elec_elec, np.size(chans_inds_iter), z_threshs,
                                        recalls[sess], rand_iteration = rand_iter)
                    for ind in range(np.size(t)):
                        key5 = [str(z) for z in z_threshs][ind]
                        output['car']['ts'][str(percent)]['elec']['elec']['kur'][key5].append(t[ind])



                    auc, auc_rand = multivar_ep(powers_mono_multi_elec, z_stds_mono_elec_ep, np.size(chans_inds_iter), z_threshs,
                                                        recalls[sess], word_lists[sess], rand_iteration = rand_iter)
                    for ind in range(np.size(auc)):
                        key5 = [str(z) for z in z_threshs][ind]
                        output['car']['aucs'][str(percent)]['elec']['epoch']['std'][key5].append(auc[ind])



                    auc, auc_rand = multivar_ep(powers_mono_multi_elec, z_kurs_mono_elec_ep, np.size(chans_inds_iter), z_threshs,
                                                        recalls[sess], word_lists[sess], rand_iteration = rand_iter)
                    for ind in range(np.size(auc)):
                        key5 = [str(z) for z in z_threshs][ind]
                        output['car']['aucs'][str(percent)]['elec']['epoch']['kur'][key5].append(auc[ind])



                    t, t_rand = univar_ep(powers_mono_uni_elec, z_stds_mono_elec_ep, np.size(chans_inds_iter), z_threshs,
                                        recalls[sess], rand_iteration = rand_iter)
                    for ind in range(np.size(t)):
                        key5 = [str(z) for z in z_threshs][ind]
                        output['car']['ts'][str(percent)]['elec']['epoch']['std'][key5].append(t[ind])



                    t, t_rand = univar_ep(powers_mono_uni_elec, z_kurs_mono_elec_ep, np.size(chans_inds_iter), z_threshs,
                                        recalls[sess], rand_iteration = rand_iter)
                    for ind in range(np.size(t)):
                        key5 = [str(z) for z in z_threshs][ind]
                        output['car']['ts'][str(percent)]['elec']['epoch']['kur'][key5].append(t[ind])

                    #PAIRS#
                    auc, auc_rand = multivar_elec(powers_pairs_multi_elec, z_stds_pairs_elec_elec, np.size(pairs_inds_iter),
                                                          z_threshs, recalls[sess], word_lists[sess], rand_iteration = None)
                    for ind in range(np.size(auc)):
                        key5 = [str(z) for z in z_threshs][ind]
                        output['pairs']['aucs'][str(percent)]['elec']['elec']['std'][key5].append(auc[ind])



                    auc, auc_rand = multivar_elec(powers_pairs_multi_elec, z_kurs_pairs_elec_elec, np.size(pairs_inds_iter),
                                                          z_threshs, recalls[sess], word_lists[sess], rand_iteration = None)
                    for ind in range(np.size(auc)):
                        key5 = [str(z) for z in z_threshs][ind]
                        output['pairs']['aucs'][str(percent)]['elec']['elec']['kur'][key5].append(auc[ind])



                    t, t_rand = univar_elec(powers_pairs_uni_elec,z_stds_pairs_elec_elec, np.size(pairs_inds_iter), z_threshs,
                                        recalls[sess], rand_iteration = rand_iter)
                    for ind in range(np.size(t)):
                        key5 = [str(z) for z in z_threshs][ind]
                        output['pairs']['ts'][str(percent)]['elec']['elec']['std'][key5].append(t[ind])



                    t, t_rand = univar_elec(powers_pairs_uni_elec,z_kurs_pairs_elec_elec, np.size(pairs_inds_iter), z_threshs,
                                        recalls[sess], rand_iteration = rand_iter)
                    for ind in range(np.size(t)):
                        key5 = [str(z) for z in z_threshs][ind]
                        output['pairs']['ts'][str(percent)]['elec']['elec']['kur'][key5].append(t[ind])



                    auc, auc_rand = multivar_ep(powers_pairs_multi_elec, z_stds_pairs_elec_ep, np.size(pairs_inds_iter), z_threshs,
                                                        recalls[sess], word_lists[sess], rand_iteration = rand_iter)
                    for ind in range(np.size(auc)):
                        key5 = [str(z) for z in z_threshs][ind]
                        output['pairs']['aucs'][str(percent)]['elec']['epoch']['std'][key5].append(auc[ind])



                    auc, auc_rand = multivar_ep(powers_pairs_multi_elec, z_kurs_pairs_elec_ep, np.size(pairs_inds_iter), z_threshs,
                                                        recalls[sess], word_lists[sess], rand_iteration = rand_iter)
                    for ind in range(np.size(auc)):
                        key5 = [str(z) for z in z_threshs][ind]
                        output['pairs']['aucs'][str(percent)]['elec']['epoch']['kur'][key5].append(auc[ind])



                    t, t_rand = univar_ep(powers_pairs_uni_elec, z_stds_pairs_elec_ep, np.size(pairs_inds_iter), z_threshs,
                                        recalls[sess], rand_iteration = rand_iter)
                    for ind in range(np.size(t)):
                        key5 = [str(z) for z in z_threshs][ind]
                        output['pairs']['ts'][str(percent)]['elec']['epoch']['std'][key5].append(t[ind])

                    t, t_rand = univar_ep(powers_pairs_uni_elec, z_kurs_pairs_elec_ep, np.size(pairs_inds_iter), z_threshs,
                                        recalls[sess], rand_iteration = rand_iter)
                    for ind in range(np.size(t)):
                        key5 = [str(z) for z in z_threshs][ind]
                        output['pairs']['ts'][str(percent)]['elec']['epoch']['kur'][key5].append(t[ind])

                    ##SUBSAMPLE EPOCHS##
                    #MONO#
                    recalls_iter = recalls[sess][epoch_inds_iter]
                    word_lists_iter = word_lists[sess][epoch_inds_iter]

                    auc, auc_rand = multivar_elec(powers_mono_multi_ep, z_stds_mono_ep_elec, num_channels[sess],
                                                          z_threshs, recalls_iter, word_lists_iter, rand_iteration = None)
                    for ind in range(np.size(auc)):
                        key5 = [str(z) for z in z_threshs][ind]
                        output['car']['aucs'][str(percent)]['epoch']['elec']['std'][key5].append(auc[ind])



                    auc, auc_rand = multivar_elec(powers_mono_multi_ep, z_kurs_mono_ep_elec, num_channels[sess],
                                                          z_threshs, recalls_iter, word_lists_iter, rand_iteration = None)
                    for ind in range(np.size(auc)):
                        key5 = [str(z) for z in z_threshs][ind]
                        output['car']['aucs'][str(percent)]['epoch']['elec']['kur'][key5].append(auc[ind])



                    t, t_rand = univar_elec(powers_mono_uni_ep, z_stds_mono_ep_elec, num_channels[sess], z_threshs,
                                        recalls_iter, rand_iteration = rand_iter)
                    for ind in range(np.size(t)):
                        key5 = [str(z) for z in z_threshs][ind]
                        output['car']['ts'][str(percent)]['epoch']['elec']['std'][key5].append(t[ind])



                    t, t_rand = univar_elec(powers_mono_uni_ep,z_kurs_mono_ep_elec, num_channels[sess], z_threshs,
                                        recalls_iter, rand_iteration = rand_iter)
                    for ind in range(np.size(t)):
                        key5 = [str(z) for z in z_threshs][ind]
                        output['car']['ts'][str(percent)]['epoch']['elec']['kur'][key5].append(t[ind])



                    auc, auc_rand = multivar_ep(powers_mono_multi_ep, z_stds_mono_ep_ep, num_channels[sess], z_threshs,
                                                        recalls_iter, word_lists_iter, rand_iteration = rand_iter)
                    for ind in range(np.size(auc)):
                        key5 = [str(z) for z in z_threshs][ind]
                        output['car']['aucs'][str(percent)]['epoch']['epoch']['std'][key5].append(auc[ind])



                    auc, auc_rand = multivar_ep(powers_mono_multi_ep, z_kurs_mono_ep_ep, num_channels[sess], z_threshs,
                                                        recalls_iter, word_lists_iter, rand_iteration = rand_iter)
                    for ind in range(np.size(auc)):
                        key5 = [str(z) for z in z_threshs][ind]
                        output['car']['aucs'][str(percent)]['epoch']['epoch']['kur'][key5].append(auc[ind])



                    t, t_rand = univar_ep(powers_mono_uni_ep, z_stds_mono_ep_ep, num_channels[sess], z_threshs,
                                        recalls_iter, rand_iteration = rand_iter)
                    for ind in range(np.size(t)):
                        key5 = [str(z) for z in z_threshs][ind]
                        output['car']['ts'][str(percent)]['epoch']['epoch']['std'][key5].append(t[ind])



                    t, t_rand = univar_ep(powers_mono_uni_ep, z_kurs_mono_ep_ep, num_channels[sess], z_threshs,
                                        recalls_iter, rand_iteration = rand_iter)
                    for ind in range(np.size(t)):
                        key5 = [str(z) for z in z_threshs][ind]
                        output['car']['ts'][str(percent)]['epoch']['epoch']['kur'][key5].append(t[ind])

                    #PAIRS#
                    auc, auc_rand = multivar_elec(powers_pairs_multi_ep, z_stds_pairs_ep_elec, num_pairs[sess],
                                                          z_threshs, recalls_iter, word_lists_iter, rand_iteration = None)
                    for ind in range(np.size(auc)):
                        key5 = [str(z) for z in z_threshs][ind]
                        output['pairs']['aucs'][str(percent)]['epoch']['elec']['std'][key5].append(auc[ind])



                    auc, auc_rand = multivar_elec(powers_pairs_multi_ep, z_kurs_pairs_ep_elec, num_pairs[sess],
                                                          z_threshs, recalls_iter, word_lists_iter, rand_iteration = None)
                    for ind in range(np.size(auc)):
                        key5 = [str(z) for z in z_threshs][ind]
                        output['pairs']['aucs'][str(percent)]['epoch']['elec']['kur'][key5].append(auc[ind])



                    t, t_rand = univar_elec(powers_pairs_uni_ep,z_stds_pairs_ep_elec, num_pairs[sess], z_threshs,
                                        recalls_iter, rand_iteration = rand_iter)
                    for ind in range(np.size(t)):
                        key5 = [str(z) for z in z_threshs][ind]
                        output['pairs']['ts'][str(percent)]['epoch']['elec']['std'][key5].append(t[ind])



                    t, t_rand = univar_elec(powers_pairs_uni_ep,z_kurs_pairs_ep_elec, num_pairs[sess], z_threshs,
                                        recalls_iter, rand_iteration = rand_iter)
                    for ind in range(np.size(t)):
                        key5 = [str(z) for z in z_threshs][ind]
                        output['pairs']['ts'][str(percent)]['epoch']['elec']['kur'][key5].append(t[ind])



                    auc, auc_rand = multivar_ep(powers_pairs_multi_ep, z_stds_pairs_ep_ep, num_pairs[sess], z_threshs,
                                                        recalls_iter, word_lists_iter, rand_iteration = rand_iter)
                    for ind in range(np.size(auc)):
                        key5 = [str(z) for z in z_threshs][ind]
                        output['pairs']['aucs'][str(percent)]['epoch']['epoch']['std'][key5].append(auc[ind])



                    auc, auc_rand = multivar_ep(powers_pairs_multi_ep, z_kurs_pairs_ep_ep, num_pairs[sess], z_threshs,
                                                        recalls_iter, word_lists_iter, rand_iteration = rand_iter)
                    for ind in range(np.size(auc)):
                        key5 = [str(z) for z in z_threshs][ind]
                        output['pairs']['aucs'][str(percent)]['epoch']['epoch']['kur'][key5].append(auc[ind])



                    t, t_rand = univar_ep(powers_pairs_uni_ep, z_stds_pairs_ep_ep, num_pairs[sess], z_threshs,
                                        recalls_iter, rand_iteration = rand_iter)
                    for ind in range(np.size(t)):
                        key5 = [str(z) for z in z_threshs][ind]
                        output['pairs']['ts'][str(percent)]['epoch']['epoch']['std'][key5].append(t[ind])

                    t, t_rand = univar_ep(powers_pairs_uni_ep, z_kurs_pairs_ep_ep, num_pairs[sess], z_threshs,
                                        recalls_iter, rand_iteration = rand_iter)
                    for ind in range(np.size(t)):
                        key5 = [str(z) for z in z_threshs][ind]
                        output['pairs']['ts'][str(percent)]['epoch']['epoch']['kur'][key5].append(t[ind])

        for key1 in output.keys():
            for key2 in output[key1].keys():
                for key3 in output[key1][key2].keys():
                    for key4 in output[key1][key2][key3].keys():
                        for key5 in output[key1][key2][key3][key4].keys():
                            for key6 in output[key1][key2][key3][key4][key5].keys():
                                for key7 in output[key1][key2][key3][key4][key5][key6].keys():
                                    output[key1][key2][key3][key4][key5][key6][key7] = np.nanmean(output[key1][key2][key3][key4][key5][key6][key7])     
    
        np.savez(outpath+'/'+sub+'_sdkur_subsample', output=output)
    except Exception as e:
        np.savez(outpath+'/'+sub+'_sdkur_subsample_ERROR', output=e)

###BOOTSTRAPPING####
def bootstrapping(sub, subpath, outpath, iterations=1000):
"""
Performs random label permutations a set number of iterations during classification.
INPUTS: 'sub': subject name as a string.
'iterations': number of bootstrap iterations
'subpath': path to where subject data is saved
'outpath': where results should be saved out to

OUTPUT:
"""
    #DEFINE CLASSIFIER AND INITIAL TERMS
    try:
        lr_classifier = LogisticRegression(C=7.2e-4, penalty='l2', class_weight='balanced', solver='liblinear')
        permute_iter = iterations
        rand_iter = None
        MB = 'MB' #remove the M or B if you want to remove the Monopolar or Bipolar analysis
        MU = 'MU' #remove the M or U if you want to remove the Multivariate or Univariate analysis
        CE = 'CE' #remove the C or E if you want to remove the Channel or Epoch exclusion analysis
        SK = 'SK' #remove the S or K if you want to remove the Standard deviation or Kurtosis analysis
        SS_CE = 'CE' #remove the C or E if you want to remove the Channel or Epoch subsampling process
        z_threshs = [0,.5,1,1.5,2,2.5,3,'none'] #z-thresholds for variance and kurtosis

        output = {}            
        if 'M' in MB:
            output['car'] = {}
        if 'B' in MB:
            output['pairs'] = {}
        for key1 in list(output.keys()):
            if 'M' in MU:
                output[key1]['aucs'] = {}
            if 'U' in MU:
                output[key1]['ts'] = {}
            for key2 in list(output[key1].keys()):
                if 'C' in CE:
                    output[key1][key2]['elec'] = {}
                if 'E' in CE:
                    output[key1][key2]['epoch'] = {}
                for key3 in list(output[key1][key2].keys()):
                    if 'S' in SK:
                        output[key1][key2][key3]['std'] = {}
                    if 'K' in SK:
                        output[key1][key2][key3]['kur'] = {}
                    for key4 in list(output[key1][key2][key3].keys()):
                        for thresh in z_threshs:
                            output[key1][key2][key3][key4][str(thresh)] = []
                            #for rand_key in ['not_rand','rand']:
                                #output[key1][key2][key3][key4][str(thresh)][rand_key] = []

        # IMPORT SUBJECT DATA
        data = np.load(subpath+'/'+sub+'_data.npz')['data'][()]
        events = data['events']
        exps = data['exps']
        sessions = data['sessions']
        channels = data['contacts']
        num_channels = data['num_contacts']
        pairs = data['pairs']
        num_pairs = data['num_pairs']
        recalls = data['recalls']
        rec = data['rec']
        not_rec = data['not_rec']
        lists = data['lists']
        num_lists = data['num_lists']
        word_lists = data['word_lists']
        num_words_sess = data['num_words_sess']
        powers_mono_uni = data['powers_car_uni']
        powers_pairs_uni = data['powers_pairs_uni']
        powers_mono_multi = data['powers_car_multi']
        powers_pairs_multi = data['powers_pairs_multi']
        stds_mono = data['stds_car']
        stds_pairs = data['stds_pairs']
        kurs_mono = data['kurs_car']
        kurs_pairs = data['kurs_pairs']

        for sess in range(np.size(sessions)): 
            # MAKE SURE THERE ARE WORDS RECALLED IN THE SESSION
            if np.size(np.unique(recalls[sess]))<2:
                continue

            # MAKE SURE NO MISMATCH IN NUMBER OF WORDS
            if powers_mono_multi[sess].shape[0] != recalls[sess].size:
                continue

            # CHECK FOR NaNs CHANNELS/PAIRS AND REMOVE THEM
            zscored_mono = zscore(powers_mono_multi[sess], axis=0, ddof=1) # ZSCORE POWERS ALONG EPOCH
            zscored_pairs = zscore(powers_pairs_multi[sess], axis=0, ddof=1) # ZSCORE POWERS ALONG EPOCH
            chan_nans = [int(x/8) for x in np.unique(np.where(np.isnan(zscored_mono)==True)[1]) if x % 8 == 0]
            pairs_nans = [int(x/8) for x in np.unique(np.where(np.isnan(zscored_pairs)==True)[1]) if x % 8 == 0]

            # REMOVE NaNs if they exist and update appropriate metrics
            if np.size(chan_nans)>0 or np.size(pairs_nans)>0:
                num_channels[sess]-=np.size(chan_nans)
                num_pairs[sess]-=np.size(pairs_nans)
                powers_mono_uni[sess] = np.reshape(powers_mono_uni[sess][~numpy.isnan(zscored_mono)],[num_words_sess[sess],-1])
                powers_mono_multi[sess] = np.reshape(powers_mono_multi[sess][~numpy.isnan(zscored_mono)],[num_words_sess[sess],-1])
                powers_pairs_uni[sess] = np.reshape(powers_pairs_uni[sess][~numpy.isnan(zscored_pairs)],[num_words_sess[sess],-1])
                powers_pairs_multi[sess] = np.reshape(powers_pairs_multi[sess][~numpy.isnan(zscored_pairs)],[num_words_sess[sess],-1])
                zscored_mono = np.reshape(zscored_mono[~numpy.isnan(zscored_mono)],[num_words_sess[sess],-1])
                zscored_pairs = np.reshape(zscored_pairs[~numpy.isnan(zscored_pairs)],[num_words_sess[sess],-1])
                stds_mono[sess] = np.delete(stds_mono[sess], chan_nans, axis=1)
                kurs_mono[sess] = np.delete(kurs_mono[sess], chan_nans, axis=1)
                stds_pairs[sess] = np.delete(stds_pairs[sess], pairs_nans, axis=1)
                kurs_pairs[sess] = np.delete(kurs_pairs[sess], pairs_nans, axis=1)

            # CALCULATE CLASSIFIER COEFFICIENTS

            # Get exclusion metrics
            # Z-score the average standard deviation or kurtosis values along epochs or electrodes
            # *_elecs means average across electrodes before z-scoring, *_eps means average across epochs before z-scoring
            z_stds_mono_elecs = zscore(np.mean(stds_mono[sess],axis=0))
            z_stds_mono_eps = zscore(np.mean(stds_mono[sess],axis=1))
            z_kurs_mono_elecs = zscore(np.mean(kurs_mono[sess],axis=0))
            z_kurs_mono_eps = zscore(np.mean(kurs_mono[sess],axis=1))
            z_stds_pairs_elecs = zscore(np.mean(stds_pairs[sess],axis=0))
            z_stds_pairs_eps = zscore(np.mean(stds_pairs[sess],axis=1))
            z_kurs_pairs_elecs = zscore(np.mean(kurs_pairs[sess],axis=0))
            z_kurs_pairs_eps = zscore(np.mean(kurs_pairs[sess],axis=1))

            for itera in range(permute_iter):
                print(itera)
                #Permute the recall labels
                recalls[sess]=np.random.permutation(recalls[sess])
            #RUN CLASSIFIER ANALYSES AND ADD RESULTS TO OUTPUT DICTIONARY
                if 'M' in MU: #MULTIVARIATE
                    if 'C' in CE: #REMOVE CHANNELS
                        if 'S' in SK: #STANDARD DEVIATION
                            auc, auc_rand = multivar_elec(powers_mono_multi[sess], z_stds_mono_elecs, num_channels[sess],
                                                          z_threshs, recalls[sess], word_lists[sess], rand_iteration = rand_iter)
                            for ind in range(np.size(auc)):
                                key5 = list(output[key1][key2][key3][key4].keys())[ind]
                                output['car']['aucs']['elec']['std'][key5].append(auc[ind])

                        if 'K' in SK: #KURTOSIS
                            auc, auc_rand = multivar_elec(powers_mono_multi[sess], z_kurs_mono_elecs, num_channels[sess],
                                                          z_threshs, recalls[sess], word_lists[sess], rand_iteration = rand_iter)
                            for ind in range(np.size(auc)):
                                key5 = list(output[key1][key2][key3][key4].keys())[ind]
                                output['car']['aucs']['elec']['kur'][key5].append(auc[ind])

                    if 'E' in CE: #REMOVE EPOCHS
                        if 'S' in SK: #STANDARD DEVIATION
                            auc, auc_rand = multivar_ep(powers_mono_multi[sess], z_stds_mono_eps, num_channels[sess], z_threshs,
                                                        recalls[sess], word_lists[sess], rand_iteration = rand_iter)
                            for ind in range(np.size(auc)):
                                key5 = list(output[key1][key2][key3][key4].keys())[ind]
                                output['car']['aucs']['epoch']['std'][key5].append(auc[ind])

                        if 'K' in SK: #KURTOSIS
                            auc, auc_rand = multivar_ep(powers_mono_multi[sess], z_kurs_mono_eps, num_channels[sess], z_threshs,
                                                        recalls[sess], word_lists[sess], rand_iteration = rand_iter)
                            for ind in range(np.size(auc)):
                                key5 = list(output[key1][key2][key3][key4].keys())[ind]
                                output['car']['aucs']['epoch']['kur'][key5].append(auc[ind])

                if 'U' in MU: #UNIVARIATE
                    if 'C' in CE: #REMOVE CHANNELS
                        if 'S' in SK: #STANDARD DEVIATION
                            t, t_rand = univar_elec(powers_mono_uni[sess],z_stds_mono_elecs, num_channels[sess], z_threshs,
                                        recalls[sess], rand_iteration = rand_iter)
                            for ind in range(np.size(t)):
                                key5 = list(output[key1][key2][key3][key4].keys())[ind]
                                output['car']['ts']['elec']['std'][key5].append(t[ind])

                        if 'K' in SK: #KURTOSIS
                            t, t_rand = univar_elec(powers_mono_uni[sess], z_kurs_mono_elecs, num_channels[sess], z_threshs,
                                        recalls[sess], rand_iteration = rand_iter)
                            for ind in range(np.size(t)):
                                key5 = list(output[key1][key2][key3][key4].keys())[ind]
                                output['car']['ts']['elec']['kur'][key5].append(t[ind])

                    if 'E' in CE: #REMOVE EPOCHS
                        if 'S' in SK: #STANDARD DEVIATION
                            t, t_rand = univar_ep(powers_mono_uni[sess], z_stds_mono_eps, num_channels[sess], z_threshs,
                                        recalls[sess], rand_iteration = rand_iter)
                            for ind in range(np.size(t)):
                                key5 = list(output[key1][key2][key3][key4].keys())[ind]
                                output['car']['ts']['epoch']['std'][key5].append(t[ind])

                        if 'K' in SK: #KURTOSIS
                            t, t_rand = univar_ep(powers_mono_uni[sess], z_kurs_mono_eps, num_channels[sess], z_threshs,
                                        recalls[sess], rand_iteration = rand_iter)
                            for ind in range(np.size(t)):
                                key5 = list(output[key1][key2][key3][key4].keys())[ind]
                                output['car']['ts']['epoch']['kur'][key5].append(t[ind])

                #BIPOLAR ANALYSIS
                if "B" in MB:
                    if 'M' in MU: #MULTIVARIATE 
                        #RUN ANALYSES AND ADD RESULTS TO OUTPUT DICTIONARY
                        if 'C' in CE: #REMOVE CHANNELS
                            if 'S' in SK: #STANDARD DEVIATION
                                auc, auc_rand = multivar_elec(powers_pairs_multi[sess], z_stds_pairs_elecs, num_pairs[sess], 
                                             z_threshs, recalls[sess], word_lists[sess], rand_iteration = rand_iter)
                                for ind in range(np.size(auc)):
                                    key5 = list(output[key1][key2][key3][key4].keys())[ind]
                                    output['pairs']['aucs']['elec']['std'][key5].append(auc[ind])

                            if 'K' in SK: #KURTOSIS
                                auc, auc_rand = multivar_elec(powers_pairs_multi[sess], z_kurs_pairs_elecs, num_pairs[sess], 
                                             z_threshs, recalls[sess], word_lists[sess], rand_iteration = rand_iter)
                                for ind in range(np.size(auc)):
                                    key5 = list(output[key1][key2][key3][key4].keys())[ind]
                                    output['pairs']['aucs']['elec']['kur'][key5].append(auc[ind])

                        if 'E' in CE: #REMOVE EPOCHS
                            if 'S' in SK: #STANDARD DEVIATION
                                auc, auc_rand = multivar_ep(powers_pairs_multi[sess], z_stds_pairs_eps, num_pairs[sess],
                                            z_threshs, recalls[sess], word_lists[sess], rand_iteration = rand_iter)
                                for ind in range(np.size(auc)):
                                    key5 = list(output[key1][key2][key3][key4].keys())[ind]
                                    output['pairs']['aucs']['epoch']['std'][key5].append(auc[ind])

                            if 'K' in SK: #KURTOSIS
                                auc, auc_rand = multivar_ep(powers_pairs_multi[sess], z_kurs_pairs_eps, num_pairs[sess],
                                            z_threshs, recalls[sess], word_lists[sess], rand_iteration = rand_iter)
                                for ind in range(np.size(auc)):
                                    key5 = list(output[key1][key2][key3][key4].keys())[ind]
                                    output['pairs']['aucs']['epoch']['kur'][key5].append(auc[ind])

                    if 'U' in MU: # UNIVARIATE
                        if 'C' in CE: #REMOVE CHANNELS
                            if 'S' in SK: #STANDARD DEVIATION
                                t, t_rand = univar_elec(powers_pairs_uni[sess], z_stds_pairs_elecs, num_pairs[sess],
                                            z_threshs, recalls[sess], rand_iteration = rand_iter)
                                for ind in range(np.size(t)):
                                    key5 = list(output[key1][key2][key3][key4].keys())[ind]
                                    output['pairs']['ts']['elec']['std'][key5].append(t[ind])

                            if 'K' in SK: #KURTOSIS
                                t, t_rand = univar_elec(powers_pairs_uni[sess], z_kurs_pairs_elecs, num_pairs[sess],
                                            z_threshs, recalls[sess], rand_iteration = rand_iter)
                                for ind in range(np.size(t)):
                                    key5 = list(output[key1][key2][key3][key4].keys())[ind]
                                    output['pairs']['ts']['elec']['kur'][key5].append(t[ind])

                        if 'E' in CE: #REMOVE EPOCHS
                            if 'S' in SK: #STANDARD DEVIATION
                                t, t_rand = univar_ep(powers_pairs_uni[sess], z_stds_pairs_eps, num_pairs[sess],
                                            z_threshs, recalls[sess], rand_iteration = rand_iter)
                                for ind in range(np.size(t)):
                                    key5 = list(output[key1][key2][key3][key4].keys())[ind]
                                    output['pairs']['ts']['epoch']['std'][key5].append(t[ind])

                            if 'K' in SK: #KURTOSIS
                                t, t_rand = univar_ep(powers_pairs_uni[sess], z_kurs_pairs_eps, num_pairs[sess],
                                            z_threshs, recalls[sess], rand_iteration = rand_iter)
                                for ind in range(np.size(t)):
                                    key5 = list(output[key1][key2][key3][key4].keys())[ind]
                                    output['pairs']['ts']['epoch']['kur'][key5].append(t[ind])

        # SAVE OUT DATA
        np.savez(outpath+'/'+sub+'_stdkur_bootstrap', output=output)

    except Exception as e:
        np.savez(outpath+'/'+sub+'_stdkur_bootsrap_error', output=e)

###################################### CLASSIFIER CONVENIENCE FUNCTIONS ######################################################
def multivar_elec(powers, metrics, num_elecs, z_threshs, recalls_sess, word_lists, rand_iteration = None):

    """ Calculates AUC when removing electrodes at various thresholds.

        Inputs: 'powers': power matrix as gathered from subject information
                'metric': standard deviation/kurtosis matrix
                'num_elecs': number of channels/pairs in the time series
                'z_threshs': z-thresholds used to exclude data. Include the term 'none' in the array to not exclude any data.
                'recalls_sess': recall vector for the session. 'word_lists' - array showing which list each epoch belongs to. 'rand_iteration' number of times to iterate through random removal. If left blank, no random analysis will be performed.
        Outputs: 'auc' - array of aucs, one entry per threshold. 'auc_rand' - array of random aucs, one entry per threshold."""

    num_freqs = int(np.shape(powers)[1]/num_elecs) # NUMBER OF FREQUENCY BINS
    num_words = int(np.shape(powers)[0]) # NUMBER OF WORDS
    # INITIALIZE OUTPUT VARIABLES
    aucs = []
    aucs_rand = []
    for thresh in z_threshs:
        try:
            if thresh == 'none': # USE ALL DATA
                keeps = range(num_elecs)
                powers_kept = np.asarray(powers)
            else: 
                keeps = np.where(metrics < thresh)[0]
                kept_indices = []
                for kept in keeps:
                    kept_indices.extend(range(kept*num_freqs,kept*num_freqs+num_freqs))
                #print(kept_indices)
                powers_kept = np.asarray(powers)[:,kept_indices]

            # MULTIVARIATE
            zscored = zscore(powers_kept, axis=0, ddof=1) # ZSCORE POWERS ALONG EPOCH
            recall_probs = []
            for l_val in np.unique(word_lists): 
                l_train = list(np.unique(word_lists)) #GET TRAINING AND VALIDATION DATA
                l_train.remove(l_val)
                val_mask=(word_lists==l_val)
                train_mask=(val_mask==False)
                rec_val=np.asarray(recalls_sess)[val_mask]
                rec_train=np.asarray(recalls_sess)[train_mask]
                zscored_val=zscored[val_mask,:]
                zscored_train=zscored[train_mask,:]
                lr_classifier.fit(zscored_train, rec_train) # FIT AND EVALUTATE CLASSIFIER
                validation_recall_prob_array = lr_classifier.predict_proba(zscored_val)[:,1]
                recall_probs.extend(validation_recall_prob_array)
            recall_probs_rs = np.reshape(recall_probs, np.size(recall_probs)) #RESHAPE FOR ROC FUNCTION TO ALLOW
            auc=roc_auc_score(recalls_sess, recall_probs_rs) # CALCULATE ROC
            aucs.append(auc)
        except:
            aucs.append(np.nan)
        
        auc_rand = []
        if rand_iteration != None: #RANDOM ANALYSIS
            for iteration in range(rand_iteration):
                try:
                    if thresh == 'none':
                        random_keeps = keeps
                        powers_kept = np.asarray(powers)
                    else:
                        random_keeps = random.sample(range(num_elecs), np.size(keeps))
                        kept_indices = []
                        for kept in random_keeps:
                            kept_indices.extend(np.arange(kept*num_freqs,kept*num_freqs+num_freqs))
                        powers_kept = np.asarray(powers)[:,kept_indices]
                    zscored = zscore(powers_kept, axis=0, ddof=1)
                    recall_probs = []
                    for l_val in np.unique(word_lists):
                        l_train = list(np.unique(word_lists))
                        l_train.remove(l_val)
                        val_mask=(word_lists==l_val)
                        train_mask=(val_mask==False)
                        rec_val=np.asarray(recalls_sess)[val_mask]
                        rec_train=np.asarray(recalls_sess)[train_mask]
                        zscored_val=zscored[val_mask,:]
                        zscored_train=zscored[train_mask,:]
                        lr_classifier.fit(zscored_train, rec_train)
                        validation_recall_prob_array = lr_classifier.predict_proba(zscored_val)[:,1]
                        recall_probs.extend(validation_recall_prob_array)
                    recall_probs_rs = np.reshape(recall_probs, np.size(recall_probs))
                    auc_rand.append(roc_auc_score(recalls_sess, recall_probs_rs))
                except:
                    auc_rand.append(np.nan)
            aucs_rand.append(np.nanmean(auc_rand))
    return(aucs, aucs_rand)

def multivar_ep(powers, metrics, num_elecs, z_threshs, recalls_sess, word_lists, rand_iteration = None):
    
    """ Calculated AUC when removing epochs at various thresholds.
        Inputs: 'powers' - power matrix. 'num_elecs' - number of channels/pairs in the time series. 'z_threshs' - z-thresholds used to exclude data. Include the term 'all' in the array to not exclude any data that iteration. 'recalls_sess' - recall vector for the session. 'word_lists' - array showing which list each epoch belongs to. 'rand_iteration' number of times to iterate through random removal. If left blank, no random analysis will be performed.
        Outputs: 'auc' - array of aucs, one entry per threshold. 'auc_rand' - array of random aucs, one entre per threshold."""

    num_freqs = int(np.shape(powers)[1]/num_elecs) # NUMBER OF FREQUENCY BINS
    num_words = int(np.shape(powers)[0]) # NUMBER OF WORDS
    #INITIALZE OUTPUT VARIABLES
    aucs = []
    aucs_rand = []
    #FIND INDICES OF EPOCHS TO REMOVE AND RECREATE POWER MATRIX
    for thresh in z_threshs:
        try:
            if thresh == 'none': # USE ALL DATA
                keeps = range(np.size(recalls_sess))
                powers_kept = np.asarray(powers)
            else: 
                keeps = np.where(metrics < thresh)[0]
                powers_kept = np.asarray(powers)[keeps,:]
            recalls_thresh = np.asarray(recalls_sess)[keeps]
            word_lists_thresh = np.asarray(word_lists)[keeps]
            zscored = zscore(powers_kept, axis=0, ddof=1) # ZSCORE POWERS ALONG EPOCH
            recall_probs = []
            for l_val in np.unique(word_lists_thresh): 
                l_train = list(np.unique(word_lists_thresh)) #GET TRAINING AND VALIDATION DATA
                l_train.remove(l_val)
                val_mask=(word_lists_thresh==l_val)
                train_mask=(val_mask==False)
                rec_val=np.asarray(recalls_thresh)[val_mask]
                rec_train=np.asarray(recalls_thresh)[train_mask]
                zscored_val=zscored[val_mask,:]
                zscored_train=zscored[train_mask,:]
                lr_classifier.fit(zscored_train, rec_train) # FIT AND EVALUTATE CLASSIFIER
                validation_recall_prob_array = lr_classifier.predict_proba(zscored_val)[:,1]
                recall_probs.extend(validation_recall_prob_array)
            recall_probs_rs = np.reshape(recall_probs, np.size(recall_probs)) #RESHAPE FOR ROC FUNCTION TO ALLOW
            auc=roc_auc_score(recalls_thresh, recall_probs_rs) # CALCULATE ROC
            aucs.append(auc)
        except:
            aucs.append(np.nan)
        
        auc_rand = []
        if rand_iteration != None: #RANDOM ANALYSIS
            for iteration in range(rand_iteration):
                try:
                    if thresh == 'none':
                        random_keeps = keeps
                        powers_kept = np.asarray(powers)
                    else:
                        random_keeps = random.sample(range(np.size(recalls_sess)), np.size(keeps))
                        random_keeps.sort()
                        powers_kept = np.asarray(powers)[random_keeps, :]
                    word_lists_thresh = word_lists[random_keeps]
                    zscored = zscore(powers_kept, axis=0, ddof=1)
                    recall_probs = []
                    for l_val in np.unique(word_lists_thresh):
                        l_train = list(np.unique(word_lists_thresh))
                        l_train.remove(l_val)
                        val_mask=(word_lists_thresh==l_val)
                        train_mask=(val_mask==False)
                        rec_val=np.asarray(recalls_sess[random_keeps])[val_mask]
                        rec_train=np.asarray(recalls_sess[random_keeps])[train_mask]
                        zscored_val=zscored[val_mask,:]
                        zscored_train=zscored[train_mask,:]
                        lr_classifier.fit(zscored_train, rec_train)
                        validation_recall_prob_array = lr_classifier.predict_proba(zscored_val)[:,1]
                        recall_probs.extend(validation_recall_prob_array)
                    recall_probs_rs = np.reshape(recall_probs, np.size(recall_probs))
                    auc_rand.append(roc_auc_score(recalls_sess[random_keeps], recall_probs_rs))
                except:
                    auc_rand.append(np.nan)
            aucs_rand.append(np.nanmean(auc_rand))
    return aucs, aucs_rand

def univar_elec(powers, metrics, num_elecs, z_threshs, recalls_sess, rand_iteration = None):

    """ Calculated t-stats when removing electrodes at various thresholds.
        Inputs: 'powers' - power matrix. 'num_elecs' - number of channels/pairs in the time series. 'z_threshs' - z-thresholds used to exclude data. Include the term 'all' in the array to not exclude any data that iteration. 'recalls_sess' - recall vector for the session. 'rand_iteration' number of times to iterate through random removal. If left blank, no random analysis will be performed.
        Outputs: 'ts' - array of t-stats, one entry per threshold. 'ts_rand' - array of random t-stats, one entre per threshold."""

    num_freqs = int(np.shape(powers)[1]/num_elecs) # NUMBER OF FREQUENCY BINS
    num_words = int(np.shape(powers)[0]) # NUMBER OF WORDS
    rec = np.where(np.asarray(recalls_sess) == 1)[0]
    not_rec = np.where(np.asarray(recalls_sess) ==0)[0]
    ts = []
    ts_rand = []
    #FIND INDICES OF ELECTRODES/EPOCHS TO REMOVE AND RECREATE POWER MATRIX
    for thresh in z_threshs:
        try:
            t_stats = []
            if thresh == 'none': # USE ALL DATA
                keeps = range(num_elecs)
                powers_kept = np.asarray(powers)
            else: 
                keeps = np.where(metrics < thresh)[0]
                kept_indices = []
                for kept in keeps:
                    kept_indices.extend(range(kept*num_freqs,kept*num_freqs+num_freqs))
                powers_kept = np.asarray(powers)[:,kept_indices]
            zscore_powers = zscore(powers_kept, axis=0, ddof=1)
            zscore_mean_powers = np.mean(zscore_powers.reshape(-1,num_freqs),axis=1).reshape(np.size(recalls_sess),np.size(keeps))
            for elec in range(np.size(keeps)):
                powers_rec = zscore_mean_powers[rec, elec]
                powers_rec_rs = np.reshape(powers_rec, np.size(powers_rec))
                powers_not_rec = zscore_mean_powers[not_rec, elec]
                powers_not_rec_rs = np.reshape(powers_not_rec, np.size(powers_not_rec))
                t_stats.append(scipy.stats.ttest_ind(powers_rec_rs, powers_not_rec_rs, equal_var=False).statistic)
            ts.append(np.mean(t_stats))
        except:
            ts.append(np.nan)
            
        t_stats_rand = []
        if rand_iteration != None: #RANDOM ANALYSIS
            for iteration in range(rand_iteration):
                try:
                    t_stats_iter = []
                    if thresh == 'none':
                        random_keeps = keeps
                        powers_kept = np.asarray(powers)
                    else:
                        random_keeps = random.sample(range(num_elecs), np.size(keeps))
                        random_keeps.sort()
                        kept_indices = []
                        for kept in random_keeps:
                            kept_indices.extend(np.arange(kept*num_freqs,kept*num_freqs+num_freqs))
                        powers_kept = np.asarray(powers)[:,kept_indices]
                    zscore_powers = zscore(powers_kept, axis=0, ddof=1)
                    zscore_mean_powers = np.mean(zscore_powers.reshape(-1,num_freqs),axis=1).reshape(np.size(recalls_sess),np.size(keeps))
                    for elec in range(np.size(keeps)):
                        powers_rec = zscore_mean_powers[rec, elec]
                        powers_rec_rs = np.reshape(powers_rec, np.size(powers_rec))
                        powers_not_rec = zscore_mean_powers[not_rec, elec]
                        powers_not_rec_rs = np.reshape(powers_not_rec, np.size(powers_not_rec))
                        t_stats_iter.append(scipy.stats.ttest_ind(powers_rec_rs, powers_not_rec_rs, equal_var=False).statistic)
                    t_stats_rand.append(np.mean(t_stats_iter))
                except:
                    t_stats_rand.append(np.nan)
            ts_rand.append(np.nanmean(t_stats_rand))
    return ts, ts_rand

def univar_ep(powers, metrics, num_elecs, z_threshs, recalls_sess, rand_iteration = None):

    """ Calculated t-stats when removing electrodes at various thresholds.
        Inputs: 'powers' - power matrix. 'num_elecs' - number of channels/pairs in the time series. 'z_threshs' - z-thresholds used to exclude data. Include the term 'all' in the array to not exclude any data that iteration. 'recalls_sess' - recall vector for the session. 'rand_iteration' number of times to iterate through random removal. If left blank, no random analysis will be performed.
        Outputs: 'ts' - array of t-stats, one entry per threshold. 'ts_rand' - array of random t-stats, one entre per threshold."""

    num_freqs = int(np.shape(powers)[1]/num_elecs) # NUMBER OF FREQUENCY BINS
    num_words = int(np.shape(powers)[0]) # NUMBER OF WORDS
    # INITIAILZE OUTPUT VARIABELS
    ts = []
    ts_rand = []
    #FIND INDICES OF ELECTRODES/EPOCHS TO REMOVE AND RECREATE POWER MATRIX
    for thresh in z_threshs:
        try:
            t_stats = []
            if thresh == 'none': # USE ALL DATA
                keeps = range(np.size(recalls_sess))
                powers_kept = np.asarray(powers)
                rec_thresh = np.asarray(recalls_sess)
            else: 
                keeps = np.where(metrics < thresh)[0]
                powers_kept = np.asarray(powers)[keeps,:]
                rec_thresh = np.asarray([recalls_sess[x] for x in keeps])
            rec = np.where(rec_thresh == 1)[0]
            not_rec = np.where(rec_thresh == 0)[0]
            zscore_powers = zscore(powers_kept, axis=0, ddof=1)
            zscore_mean_powers = np.mean(zscore_powers.reshape(-1,num_freqs),axis=1).reshape(np.size(keeps),num_elecs)
            for elec in range(num_elecs):
                powers_rec = zscore_mean_powers[rec, elec]
                powers_rec_rs = np.reshape(powers_rec, np.size(powers_rec))
                powers_not_rec = zscore_mean_powers[not_rec, elec]
                powers_not_rec_rs = np.reshape(powers_not_rec, np.size(powers_not_rec))
                t_stats.append(scipy.stats.ttest_ind(powers_rec_rs, powers_not_rec_rs, equal_var=False).statistic)
            ts.append(np.nanmean(t_stats))
        except:
            ts.append(np.nan)
        
        t_stats_rand = []
        if rand_iteration != None: #RANDOM ANALYSIS
            for iteration in range(rand_iteration):
                try:
                    t_stats_iter = []
                    if thresh == 'none':
                        powers_kept = np.asarray(powers)
                        rec_thresh = recalls_sess
                    else:
                        keeps = random.sample(range(np.size(recalls_sess)), np.size(keeps))
                        keeps.sort()
                        powers_kept = np.asarray(powers)[keeps,:]
                        rec_thresh = recalls_sess[keeps]
                    rec = np.where(rec_thresh == 1)[0]
                    not_rec = np.where(rec_thresh == 0)[0]
                    zscore_powers = zscore(powers_kept, axis=0, ddof=1)
                    zscore_mean_powers = np.mean(zscore_powers.reshape(-1,num_freqs),axis=1).reshape(np.size(keeps),num_elecs)
                    for elec in range(num_elecs):
                        powers_rec = zscore_mean_powers[rec, elec]
                        powers_rec_rs = np.reshape(powers_rec, np.size(powers_rec))
                        powers_not_rec = zscore_mean_powers[not_rec, elec]
                        powers_not_rec_rs = np.reshape(powers_not_rec, np.size(powers_not_rec))
                        t_stats_iter.append(scipy.stats.ttest_ind(powers_rec_rs, powers_not_rec_rs, equal_var=False).statistic)
                    t_stats_rand.append(np.mean(t_stats_iter))
                except:
                    t_stats_rand.append(np.nan)
            ts_rand.append(np.nanmean(t_stats_rand))
    return ts, ts_rand

def manual_elec_classification(powers_multi, powers_uni, num_elecs, inds_to_remove, recalls_sess, word_lists):

    """ Calculates AUC when removing electrodes at various thresholds.

        Inputs: 'powers_multi': multivariate powers
                'powers_uni': univariate powers
                'num_elecs': number of channels/pairs in the time series
                'inds_to_remove': Indices of electrodes to remove
                'recalls_sess': recall vector for the session. 
                'word_lists' - array showing which list each epoch belongs to.
        Outputs: 'auc' - array of aucs, one entry per threshold. 't' - array of t-stats, one entry per threshold."""

    # MULTIVAR
    num_freqs = int(np.shape(powers_multi)[1]/num_elecs) # NUMBER OF FREQUENCY BINS
    num_words = int(np.shape(powers_multi)[0]) # NUMBER OF WORDS
    keeps = list(np.arange(num_elecs))
    for ind_to_remove in inds_to_remove:
        keeps.remove(ind_to_remove)
    kept_indices=[]
    for kept in keeps:
        kept_indices.extend(range(kept*num_freqs,kept*num_freqs+num_freqs))
    powers_kept_multi = np.asarray(powers_multi)[:,kept_indices]
    powers_kept_uni = np.asarray(powers_uni)[:,kept_indices]
    zscored_multi = zscore(powers_kept_multi, axis=0, ddof=1) # ZSCORE POWERS ALONG EPOCH
    zscored_uni = zscore(powers_kept_uni, axis=0, ddof=1) # ZSCORE POWERS ALONG EPOCH
    try:
        recall_probs = []
        for l_val in np.unique(word_lists): 
            l_train = list(np.unique(word_lists)) #GET TRAINING AND VALIDATION DATA
            l_train.remove(l_val)
            val_mask=(word_lists==l_val)
            train_mask=(val_mask==False)
            rec_val=np.asarray(recalls_sess)[val_mask]
            rec_train=np.asarray(recalls_sess)[train_mask]
            zscored_val=zscored_multi[val_mask,:]
            zscored_train=zscored_multi[train_mask,:]
            lr_classifier.fit(zscored_train, rec_train) # FIT AND EVALUTATE CLASSIFIER
            validation_recall_prob_array = lr_classifier.predict_proba(zscored_val)[:,1]
            recall_probs.extend(validation_recall_prob_array)
        recall_probs_rs = np.reshape(recall_probs, np.size(recall_probs)) #RESHAPE FOR ROC FUNCTION TO ALLOW
        auc=roc_auc_score(recalls_sess, recall_probs_rs) # CALCULATE ROC
    except:
        auc = np.nan
        
    #UNIVAR
    try:
        t_stats=[]
        rec = np.where(np.asarray(recalls_sess) == 1)[0]
        not_rec = np.where(np.asarray(recalls_sess) ==0)[0]
        #FIND INDICES OF ELECTRODES/EPOCHS TO REMOVE AND RECREATE POWER MATRIX
        zscore_mean_powers = np.mean(zscored_uni.reshape(-1,num_freqs),axis=1).reshape(np.size(recalls_sess),np.size(keeps))
        for elec in range(np.size(keeps)):
            powers_rec = zscore_mean_powers[rec, elec]
            powers_rec_rs = np.reshape(powers_rec, np.size(powers_rec))
            powers_not_rec = zscore_mean_powers[not_rec, elec]
            powers_not_rec_rs = np.reshape(powers_not_rec, np.size(powers_not_rec))
            t_stats.append(scipy.stats.ttest_ind(powers_rec_rs, powers_not_rec_rs, equal_var=False).statistic)
        t = (np.mean(t_stats))
    except Exception as e:
        t = np.nan
    return(auc, t)